/* $Id: extract_candy.c,v 1.22 2012/04/06 21:37:46 myself Exp $
 *
 * extract_candy.c  -  Extract lineations from Cande et al., 1989, Magnetic lineations
 * of the world's ocean basins, AAPG, Tulsa, OK.
 *
 * Geoware, (c) 2000-2012 P. Wessel
 * 9/26/01 - Now with byte swabing
 * 10/26/04 - geoware_companion include file
 * 4/10/11 - geoware.h include file
 */
 
#include "geoware.h"

#define N_F 29
#define N_O 4
#define N_P 50

#define FZFILE  "cande_fz_and_features.b"
#define MAGFILE "cande_mag_lineations.b"

char *text_feature[N_F] = {
	"ISOCHRON",
	"RIDGE",
	"BASIN",
	"ISLAND",
	"ISLANDS",
	"TRENCH",
	"FRACTURE_ZONE",
	"TROUGH",
	"BANK",
	"ARCHIPELAGO",
	"PLATEAU",
	"RISE",
	"SEAMOUNT",
	"SEAMOUNTS",
	"SEA",
	"ABYSSAL_PLAIN",
	"SEAMOUNT_CHAIN",
	"DEEP",
	"RIDGE_COMPLEX",
	"MAGNETIC_ANOMALY",
	"GUYOT",
	"PASSAGE",
	"SCARP",
	"FAN",
	"ESCARPMENT",
	"TRIPLE_JUNCTION",
	"SPREADING_DISCONTINUITY",
	"FRACTURE",
	"FOSSIL_RIDGE"
};

char *text_ocean[N_O] = {
	"Pacific",
	"North-Atlantic",
	"South-Atlantic",
	"Indian"
};

char *text_plate[N_P] = {
	"PACIFIC [PAC]",
	"NORTH_AMERICA [NAM]",
	"EURASIA [EUR]",
	"SOUTH_AMERICA [SAM]",
	"AFRICA [AFR]",
	"NAZCA [NAZ]",
	"COCOS [COC]",
	"CARIBBEAN [CAR]",
	"ANTARCTICA [ANT]",
	"AUSTRALIA [AUS]",
	"INDIA [IND]",
	"ARABIA [ARA]",
	"PHILIPPINE [PHI]",
	"BISMARCK [BIS]",
	"NEW_BRITIAN [NBR]",
	"SOLOMON [SOL]",
	"WEST_FIJI [WFI]",
	"HAVRE-LAU [HLA]",
	"RIVERA [RIV]",
	"JUAN_DE_FUCA [JFU]",
	"SCOTIA [SCO]",
	"SOUTH_SANDWICH [SSA]",
	"SOUTH_SHETLAND [SSH]",
	"EASTER [EAS]",
	"JUAN_FERNANDEZ [JFE]",
	"BURMA [BUR]",
	"CAMPBELL_PLATEAU [CAP]",
	"CHATHAM [CHA]",
	"NORTH_NEW_ZEALAND [NNZ]",
	"LORD_HOWE_RISE [LHR]",
	"NORFOLK_RIDGE [NOR]",
	"SOUTH_TASMAN_RISE [STR]",
	"KERGUELEN_PLATEAU [KEP]",
	"MADAGASCAR [MAD]",
	"ANTARCTIC_PENINSULA [ANP]",
	"FILCHNER [FIL]",
	"HAAG [HAA]",
	"THURSTON [THU]",
	"MARIA_BYRD_LAND [MBL]",
	"ELLSWORTH [ELL]",
	"EAST_ANTARCTICA [EAN]",
	"ROSS [ROS]",
	"NUBIA [NUB]",
	"SOMALIA [SOM]",
	"MAYA [MAY]",
	"CHORTISS [CHO]",
	"NICOYA [NIC]",
	"IBERIA [IBE]",
	"W.(MARGIN)_N.AMERICA [WNA]",
	"W.(MARGIN)_S.AMERICA [WSA]",
};

int outside (float w1, float e1, float s1, float n1, float w2, float e2, float s2, float n2);

int main (int argc, char **argv)
{
	int i, feature, ocean, plate, tmp1, tmp2, tmp3, n_features, n_points, found = 0;
	int error = FALSE, list = FALSE, verbose = FALSE, individual = FALSE;
	int area_set = FALSE, ocean_set = FALSE, plate_set = FALSE, greenwich = TRUE, gmt = FALSE;
#if defined(_WIN32) || WORDS_BIGENDIAN == 0
	int *p4;
#endif
	float w, e, s, n, xy[10000][2];
	struct REC {
		int np;
		float w, e, s, n;
		short int ocean2;	/* 0 if not listed, otherwise 1 - 4 */
		short int plate2;	/* 0 if not listed, otherwise 1 - 50 */
		short int id2;
		short int feature2;
		char label[20];
	} header;
	FILE *fp = NULL, *fpo;
	char alternate[10], prefix[80], file[BUFSIZ], *directory, *dir, *p, *choice;
	
	if ( (directory = getenv ("GMT_DATADIR")) == (char *)NULL) {
		fprintf (stderr, "%s: Environmental variable GMT_DATADIR not set!\n", argv[0]);
		exit (EXIT_FAILURE);
	}
	if (sizeof (struct REC) != 48) {
		fprintf (stderr, "%s: struct REC != 48 bytes!  Notify Geoware\n", argv[0]);
		exit (EXIT_FAILURE);
	}
	feature = -1;		/* Must specify feature */
	plate = ocean = 0;	/* Defaults is all plates in all oceans */
	
	for (i = 1; i < argc; i++) {
		switch (argv[i][1]) {
			case 'D':	/* Wants individual files */
				strcpy (prefix, &argv[i][2]);
				individual = TRUE;
				break;
			case 'F':	/* Feature */
				feature = atoi (&argv[i][2]);
				break;
			case 'O':	/* Ocean */
				ocean = atoi (&argv[i][2]);
				ocean_set = TRUE;
				break;
			case 'P':	/* Plate */
				plate = atoi (&argv[i][2]);
				plate_set = TRUE;
				break;
			case 'R':	/* Region */
				sscanf (&argv[i][2], "%f/%f/%f/%f", &w, &e, &s, &n);
				area_set = TRUE;
				break;
			case 'L':
				list = error = TRUE;
				break;
			case 'G':	/* GMT format */
				gmt = TRUE;
				break;
			case 'V':
				verbose = TRUE;
				break;
			default:
				error = TRUE;
				break;
		}
	}
	
	if (feature < 0 || feature >= N_F) error = TRUE;
	if (ocean_set && (ocean < 1 || ocean > N_O)) error = TRUE;
	if (plate_set && (plate < 1 || plate > N_P)) error = TRUE;
	
	if (argc == 1 || error) {
		fprintf (stderr, "extract_candy %s - Extract lines from Cande et al., 1989 Magnetic Lineations Map\n\n", COMP_VERSION);
		fprintf (stderr, "Usage: extract_candy -Ffeature [-Dprefix -L -Oocean -Pplate -Rw/e/s/n -V]\n");
		fprintf (stderr, "	Retrieves features from Cande et al's Magnetic Lineation map.\n");
		fprintf (stderr, "	-D sets prefix for individual files [Default is stdout]\n");
		fprintf (stderr, "	-F sets the feature number [0]\n");
		fprintf (stderr, "	-L lists the features, oceans, and plates available\n");
		fprintf (stderr, "	-O sets the ocean [all oceans]\n");
		fprintf (stderr, "	-P sets the plate [all plates]\n");
		fprintf (stderr, "	-R sets the region [the world]\n");
		fprintf (stderr, "	-V to run in verbose mode\n");
		if (list) {
			fprintf (stderr, "	   The following features can be extracted (specify number):\n\n");
			for (i = 0; i < N_F; i++) fprintf (stderr,"\t\t%2d - %s\n", i, text_feature[i]);
			fprintf (stderr, "	   The following oceans can be selected:\n");
			for (i = 0; i < N_O; i++) fprintf (stderr,"\t\t%2d - %s\n", i+1, text_ocean[i]);
			fprintf (stderr, "	   The following plates can be selected:\n");
			for (i = 0; i < N_P; i++) fprintf (stderr,"\t\t%2d - %s\n", i+1, text_plate[i]);
		}
		exit (EXIT_FAILURE);
	}
	
	if (w < 0.0 && e > 0.0) greenwich = FALSE;
	if (!greenwich && w >= 180.0) {
		w -= 360.0;
		e -= 360.0;
	}
	choice = (feature == 0) ? MAGFILE : FZFILE;	/* Magnetic lineations or another feature */
	dir = directory;
	while (!found && (p = strtok (dir, PATH_SEPARATOR))) {
		sprintf (file, "%s%c%s", p, DIR_DELIM, choice);
		if (!access (file, R_OK)) found = TRUE;
		dir = NULL;
	}
	if ((fp = fopen (file, "rb")) == NULL) {
		fprintf (stderr, "extract_candy:  Cannot open file %s!\n", file);
		exit (EXIT_FAILURE);
	}
	else if (verbose)
		fprintf (stderr, "extract_candy: Extract data from %s\n", file);
	
	if (verbose) {
		if (feature >= 0)
			fprintf (stderr, "extract_candy: Retrieve %s from ", text_feature[feature]);
		else
			fprintf (stderr, "extract_candy: Retrieve all features from ");
		if (ocean_set)
			fprintf (stderr, "the %s ocean", text_ocean[ocean-1]);
		else
			fprintf (stderr, "all oceans");
		if (plate_set) fprintf (stderr, " on plate %s.", text_plate[plate-1]);
		fprintf (stderr, "...");
	}
	n_features = n_points = 0;
	
	fpo = stdout;	/* Default is 1 file sent to stdout */
	
	while (fread ((void *)&header, 48, 1, fp) == 1) {
		
#if defined(_WIN32) || WORDS_BIGENDIAN == 0
		header.np = swab_4 (header.np);
		p4 = (int *)&header.w; *p4 = swab_4 (*p4);
		p4 = (int *)&header.e; *p4 = swab_4 (*p4);
		p4 = (int *)&header.s; *p4 = swab_4 (*p4);
		p4 = (int *)&header.n; *p4 = swab_4 (*p4);
		header.ocean2 = swab_2 (header.ocean2);
		header.plate2 = swab_2 (header.plate2);
		header.id2 = swab_2 (header.id2);
		header.feature2 = swab_2 (header.feature2);
#endif
		/* Check if we need to skip this segment */
		
		if ((header.feature2 != feature) ||
			(plate_set && header.plate2 != plate && feature != 6) ||
			(ocean_set && header.ocean2 != ocean) ||
			(area_set && outside (w, e, s, n, header.w, header.e, header.s, header.n))) {
			fseek (fp, (long) (header.np * 8), 1);
			continue;
		}
			
		n_features++;
		n_points += header.np;
		
		if (individual) {
			sprintf (file, "%s_%d.d", prefix, n_features);
			if ((fpo = fopen (file, "w")) == NULL) {
				fprintf (stderr, "extract_candy: Could not create file %s\n", file);
				exit (EXIT_FAILURE);
			}
		}
		
		/* First print out segment header */
		
		if (header.feature2 == 0) {	/* Magnetic lineations */
			if (header.ocean2 > 255 || header.plate2 > 255 || header.id2 > 255) {
				tmp1 = header.ocean2 / 256;
				tmp2 = header.plate2 / 256;
				tmp3 = header.id2 / 256;
				sprintf (alternate, "%1d%2.2d%2.2d", tmp1, tmp2, tmp3);
			}
			else
				strcpy (alternate, "NONE ");
			if (gmt)
				fprintf (fpo, "> %4d OCEAN: %s PLATE: %s ID_NO: %2d ALT: %s ANOMALY_ID: %s -L\"%s\"\n",
					header.np, text_ocean[header.ocean2-1], text_plate[header.plate2-1], (int)header.id2, alternate, header.label, header.label);
			else
				fprintf (fpo, "> %4d OCEAN: %1d PLATE: %2d ID_NO: %2d ALT: %s ANOMALY_ID: %s\n",
					header.np, (int)header.ocean2, (int)header.plate2, (int)header.id2, alternate, header.label);
		}
		else {
			if (gmt)
				fprintf (fpo, "> %4d OCEAN: %s FEATURE: %s NAME: %s -L\"%s\"\n",
					header.np, text_ocean[header.ocean2-1], text_feature[header.feature2], header.label, header.label);
			else
				fprintf (fpo, "> %4d OCEAN: %1d FEATURE: %2d NAME: %s\n",
					header.np, (int)header.ocean2, (int)header.feature2, header.label);
		}
		
		fread ((void *)xy, 8, (int)header.np, fp);
		
		for (i = 0; i < header.np; i++) {
#if defined(_WIN32) || WORDS_BIGENDIAN == 0
			p4 = (int *)&xy[i][0]; *p4 = swab_4 (*p4);
			p4 = (int *)&xy[i][1]; *p4 = swab_4 (*p4);
#endif
			if (!greenwich && xy[i][0] >= 180.0) xy[i][0] -= 360.0;
			fprintf (fpo, "%.5f\t%.5f\n", xy[i][0], xy[i][1]);
		}
		
		if (individual) fclose (fpo);
	}
	
	fclose (fp);
	
	if (verbose) fprintf (stderr, "\nextract_candy: Retrieved %d features (total of %d points)\n", n_features, n_points);
	
	exit (EXIT_SUCCESS);
}

int outside (float w1, float e1, float s1, float n1, float w2, float e2, float s2, float n2)
{
	int out = 0;
	if (s1 >= n2 || n1 <= s2) return (TRUE);
	w2 -= 360.0;
	while (w2 < w1) w2 += 360.0;
	if (w2 >= e1) out = 1;
	e2 += 360.0;
	while (e2 > e1) e2 -= 360.0;
	if (e2 <= w1) out++;
	out--;
	return ( (out > 0) ? TRUE : FALSE );
}
