/*
 * $Id: extract_cities.c,v 1.20 2012/04/06 21:37:46 myself Exp $
 *
 * extract_cities.c reads the UN Population database and extracts a
 * subset of cities/capitals/countries based on the command-line selections.
 *
 * See file COPYING for Copyright information
 *
 * Geoware, (c) 2000-2012 P. Wessel
 */

#include "geoware.h"

/*
* This is the format of the tab-delimited data records:
CO	000	22720	CM	 65.0	 33.5	0.0	Afghanistan
CI	000	177300	LM	 62.17	 34.33	177300	0	Herat
CA	000	1424400	LM	 69.17	 34.50	1424400	0	KABUL
CI	000	225500	LM	 65.78	 31.60	225500	0	Kandahar (Quandahar)
*/

#define UNFILE "UN_cities.dat"

struct UN {
	double lon, lat, angle;
	int population;
	char just[4], id[4], name[32];
};

int main (int argc, char **argv) {
	int i, j, n_alloc, size = 9, n_cities = 0, population = 0, found = 0;
	int error = FALSE, text = FALSE, loc = FALSE, verbose = FALSE, got_angle = FALSE, got_id = FALSE, country = FALSE, header = FALSE;
	char line[BUFSIZ], file[BUFSIZ], *dir, *p, *directory, *just = CNULL, *id = CNULL, *font = CNULL, *get = CNULL, Helvetica[2];
	double angle = 0.0, west = 0.0, east = 360.0, south = -90.0, north = 90.0;
	FILE *fp = NULL;
	struct UN *cities;

	if ( (directory = getenv ("GMT_DATADIR")) == (char *)NULL) {
		fprintf (stderr, "%s: Environmental variable GMT_DATADIR not set!\n", argv[0]);
		exit (EXIT_FAILURE);
	}
	Helvetica[0] = '0';	Helvetica[1] = '\0';	/* Font number 0 */

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'R':
					sscanf (&argv[i][2], "%lf/%lf/%lf/%lf", &west, &east, &south, &north);
					break;
				case 'A':
					angle = atof (&argv[i][2]);
					got_angle = TRUE;
					break;
				case 'F':
					font = &argv[i][2];
					break;
				case 'I':
					id = &argv[i][2];
					got_id = TRUE;
					break;
				case 'J':
					just = &argv[i][2];
					break;
				case 'P':
					population = atoi (&argv[i][2]);
					break;
				case 'S':
					size = atoi (&argv[i][2]);
					break;
				case 'T':
					text = TRUE;
					break;
				case 'L':
					loc = TRUE;
					break;
				case 'H':
					header = TRUE;
					break;
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else
			get = argv[i];
	}

	if (argc == 1 || error) {
		fprintf (stderr, "extract_cities %s - Extract cities from UN data base\n\n", COMP_VERSION);
		fprintf (stderr, "usage: extract_cities CA|CI|CO  [-A<angle>] [-F<font>] [-I<id>] [-H] [-J<just>] [-L]\n");
		fprintf (stderr, "	[-P<population>] [-R<west/east/south/north>] [-S<fontsize>] [-T] [-V]\n\n");
		fprintf (stderr, "	Select one of CA(pital), CI(ty), or CO(untry)\n\n");
		fprintf (stderr, "	OPTIONS\n");
		fprintf (stderr, "	-A Set angle for text label [0].  Requires -T\n");
		fprintf (stderr, "	-F Set font name or number (see pstext) [Helvetica (0)].  Requires -T\n");
		fprintf (stderr, "	-H Output a 1-record header [no header]\n");
		fprintf (stderr, "	-I Only extract data for the country with this id [all countries]\n");
		fprintf (stderr, "	-J Set fixed text justification (see pstext) [Default is preset in database].  Requires -T\n");
		fprintf (stderr, "	-L Output results as input for psxy [Default]\n");
		fprintf (stderr, "	-P Only extract items with populations exceeding <population> [0]\n");
		fprintf (stderr, "	-R Set region [0/360/-90/90]\n");
		fprintf (stderr, "	-S Set font size in points [9].  Requires -T\n");
		fprintf (stderr, "	-T Output results as input for pstext\n");
		fprintf (stderr, "	-V will report progress while working [Default is quiet]\n");
		exit (EXIT_FAILURE);
	}
	
	if (!get) {
		fprintf (stderr, "%s:  Must choose among CA(pital), CI(ty), or CO(untry)!\n", argv[0]);
		exit (EXIT_FAILURE);
	}
	if (text && loc) {
		fprintf (stderr, "%s:  Must choose between -L and -T!\n", argv[0]);
		exit (EXIT_FAILURE);
	}
	if (!text) loc = TRUE;
	if (!font) font = Helvetica;
	
	dir = directory;
	while (!found && (p = strtok (dir, PATH_SEPARATOR))) {
		sprintf (file, "%s%c%s", p, DIR_DELIM, UNFILE);
		if (!access (file, R_OK)) found = TRUE;
		dir = NULL;
	}
	if ((fp = fopen (file, "r")) == NULL) {
		fprintf (stderr, "%s:  Cannot open file %s!\n", argv[0], file);
		exit (EXIT_FAILURE);
	}
	else if (verbose)
		fprintf (stderr, "%s: Extract data from %s\n", argv[0], file);
	
	n_alloc = CHUNK;
	cities = (struct UN *) malloc ((size_t)(n_alloc * sizeof (struct UN)));
	country = (!strcmp (get, "CO"));
	while (fgets (line, BUFSIZ, fp) && (line[0] == '#' || line[0] == '\n')); /* Skip header */

	i = 0;
	while (fgets (line, BUFSIZ, fp)) {
		if (line[0] == '#' || line[0] == '\n') continue;	/* Skip headers */
		if (strncmp (line, get, 2)) continue;			/* Wrong type */
		
		if (country) {
			sscanf (line, "%*s %s %d %s %lf %lf %lf %[^\n]", cities[i].id, &cities[i].population, cities[i].just, &cities[i].lon, &cities[i].lat, &cities[i].angle, cities[i].name);
			cities[i].population *= 1000;
		}
		else
			sscanf (line, "%*s %s %d %s %lf %lf %*d %*d %[^\n]", cities[i].id, &cities[i].population, cities[i].just, &cities[i].lon, &cities[i].lat, cities[i].name);
			
		if (got_id && !strcmp (cities[i].id, id)) continue;
		if (cities[i].lat < south || cities[i].lat > north) continue;
		if (cities[i].population < population) continue;
		cities[i].lon -= 360.0;
		while (cities[i].lon < west) cities[i].lon += 360.0;
		if (cities[i].lon > east) continue;
		j = 31;
		while (j > 0 && cities[i].name[j] != '(') j--;
		if (j > 0) cities[i].name[j-1] = '\0';
		while (cities[i].lon < 0.0) cities[i].lon += 360.0;
		if (just) strcpy (cities[i].just, just);
		if (got_angle) cities[i].angle = angle;
		i++;
		if (i == n_alloc) {
			n_alloc += CHUNK;
			cities = (struct UN *) realloc ((void *)cities, (size_t)(n_alloc * sizeof (struct UN)));
		}
	}
	n_cities = i;
	fclose (fp);
	
	if (verbose) fprintf (stderr, "%s: Found %d cities matching your search criteria\n", argv[0], n_cities);
	
	if (!n_cities) {
		free ((void *)cities);
		exit (EXIT_SUCCESS);
	}
	
	if (text) {	/* pstext format output */
		if (header) printf ("#lon\tlat\tfsize\tangle\tfont\tjust\ttext\n");
		for (i = 0; i < n_cities; i++) printf ("%g\t%g\t%d\t%g\t%s\t%s\t%s\n", cities[i].lon, cities[i].lat, size, cities[i].angle, font, cities[i].just, cities[i].name);
	}
	else {	/* Locations only */
		if (header) printf ("#lon\tlat\tpopulation\n");
		for (i = 0; i < n_cities; i++) printf ("%g\t%g\t%d\n", cities[i].lon, cities[i].lat, cities[i].population);
	}
		
	free ((void *)cities);
	if (verbose) fprintf (stderr, "%s: Done\n", argv[0]);
	
	exit (EXIT_SUCCESS);
}
