/*
 * $Id: extract_countries.c,v 1.11 2012/04/06 21:37:46 myself Exp $
 *
 * extract_countries.c takes the given country code and return a fillable polygon
 * file that can be plotted with psxy.  The data used is the Digital Chart of
 * the World (DCW), processed by Geoware to yield GMT multi-segment files.
 *
 * See file COPYING for Copyright information
 *
 * Geoware, (c) 2000-2012 P. Wessel
 */

#include "geoware.h"

#define DCW_DIR			"DCW"
#define DCW_COUNTRIES		230
#define DCW_STATES		97
#define N_CONTINENTS		8
#define N_COUNTRIES_WITH_STATES	4
struct COUNTRY {
	char continent[3];
	char code[3];
	char name[80];
};

struct STATE {
	char country[3];
	char code[3];
	char name[80];
};

struct COUNTRY country[DCW_COUNTRIES] = {
#include "geoware_dcw.h"
};
struct STATE states[DCW_STATES] = {
#include "geoware_states.h"
};

char *continents[N_CONTINENTS] = {"Africa", "Antarctica", "Asia", "Europe", "Oceania", "North America", "South America", "Miscellaneous"};
char *country_with_states[N_COUNTRIES_WITH_STATES] = {"AU", "BR", "CA", "US"};

int find_country (char *code, struct COUNTRY *list, int n);
int find_state (char *scode, char *ccode, struct STATE *slist, int ns);
int country_has_states (char *code, char *countries[], int n);

int main (int argc, char **argv) {
	int i, j, k, error = FALSE, verbose = FALSE, listing = FALSE, list_states = FALSE, want_state = FALSE, found = FALSE;
	int comp_countries (const void *p1, const void *p2);

	char line[BUFSIZ], file[BUFSIZ], code[64], state[64], *directory, *dir, *p;

	FILE *fp = NULL;

	if ( (directory = getenv ("GMT_DATADIR")) == (char *)NULL) {
		fprintf (stderr, "%s: Environmental variable GMT_DATADIR not set!\n", argv[0]);
		exit (EXIT_FAILURE);
	}

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'L':
					listing = TRUE;
					if (argv[i][2] == 's') list_states = TRUE;
					break;
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
	}

	if (argc == 1 || error) {
		fprintf (stderr, "extract_countries %s - Extract country polygons from DCW data base of %d countries\n\n", COMP_VERSION, DCW_COUNTRIES);
		fprintf (stderr, "usage: extract_countries code1 code2 ... [-L[s]] [-V] > polygons.dat\n\n");
		fprintf (stderr, "	code1, code2 ,etc are 2-character country codes\n");
		fprintf (stderr, "	To get a state of a country (if present), append .state, e.g, US.TX for Texas.\n\n");
		fprintf (stderr, "	OPTIONS\n");
		fprintf (stderr, "	-L Display list of countries and their codes\n");
		fprintf (stderr, "	   Append s to see states/terretories for Australia, Brazil, Canada, and the US.\n");
		fprintf (stderr, "	-V will report progress while working [Default is quiet]\n");
		exit (EXIT_FAILURE);
	}
	
	if (listing) {
		for (i = k = 0; i < DCW_COUNTRIES; i++) {
			if (i == 0 || strcmp (country[i].continent, country[i-1].continent) ) {
				printf ("%s:\n", continents[k]);
				k++;
			}
			printf ("  %s\t%s\t%s\n", country[i].continent, country[i].code, country[i].name);
			if (list_states && country_has_states (country[i].code, country_with_states, N_COUNTRIES_WITH_STATES)) {
				for (j = 0; j < DCW_STATES; j++) {
					if (!strcmp (country[i].code, states[j].country)) printf ("\t\t%s.%s\t%s\n", country[i].code, states[j].code, states[j].name);
				}
			}
		}
		exit (EXIT_SUCCESS);
	}

	qsort ((void *)country, (size_t)DCW_COUNTRIES, sizeof (struct COUNTRY), comp_countries);	/* Sort on country code */
 	
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') continue;	/* Skip options this time */
		
		memset ((void *)code, 0, 64);
		strcpy (code, argv[i]);
		want_state = FALSE;
		if (code[2] == '.') {	/* Requesting a state */
			memset ((void *)state, 0, 64);
			strcpy (state, &code[3]);
			code[2] = '\0';
			want_state = TRUE;
		}
		k = find_country (code, country, DCW_COUNTRIES);
		if (k == -1) {
			fprintf (stderr, "%s:  No country code matching %s (skipped)\n", argv[0], code);
			continue;
		}
		dir = directory;
		if (want_state) {
			if ((j = find_state (state, code, states, DCW_STATES)) == -1) {
				fprintf (stderr, "%s:  Country %s does not have states (skipped)\n", argv[0], code);
				continue;
			}
			while (!found && (p = strtok (dir, PATH_SEPARATOR))) {
				sprintf (file, "%s%c%s%c%s%c%s%c%s.dat", p, DIR_DELIM, DCW_DIR, DIR_DELIM, country[k].continent, DIR_DELIM, country[k].code, DIR_DELIM, states[j].code);
				if (!access (file, R_OK)) found = TRUE;
				dir = NULL;
			}
		}
		else {
			while (!found && (p = strtok (dir, PATH_SEPARATOR))) {
				sprintf (file, "%s%c%s%c%s%c%s.dat", p, DIR_DELIM, DCW_DIR, DIR_DELIM, country[k].continent, DIR_DELIM, country[k].code);
				if (!access (file, R_OK)) found = TRUE;
				dir = NULL;
			}
		}
			
		if ((fp = fopen (file, "r")) == NULL) {
			fprintf (stderr, "%s:  Cannot open file %s!\n", argv[0], file);
			exit (EXIT_FAILURE);
		}
		else if (verbose) {
			if (want_state) 
				fprintf (stderr, "%s: Extract data for %s (%s)\n", argv[0], states[j].name, country[k].name);
			else
				fprintf (stderr, "%s: Extract data for %s\n", argv[0], country[k].name);
		}
		
		while (fgets (line, BUFSIZ, fp)) {	/* Just dumb copy */
			printf ("%s", line);
		}
		fclose (fp);
	}
			
	if (verbose) fprintf (stderr, "%s: Done\n", argv[0]);
	
	exit (EXIT_SUCCESS);
}

int comp_countries (const void *p1, const void *p2)
{
	struct COUNTRY *A, *B;
	A = (struct COUNTRY *)p1;
	B = (struct COUNTRY *)p2;
	return (strcmp (A->code, B->code));
}

int find_country (char *code, struct COUNTRY *list, int n)
{
	int low, high, mid, last = -1, way;
	
	low = 0;
	high = n;
	while (low < high) {
		mid = (low + high) / 2;
		if (mid == last) return (-1);	/* No such code */
		way = strcmp (code, list[mid].code);
		if (way > 0)
			low = mid;
		else if (way < 0)
			high = mid;
		else 
			return (mid);
		last = mid;
	}
	return (low);
}

int find_state (char *scode, char *ccode, struct STATE *slist, int ns)
{
	int i;
	
	for (i = 0; i < ns; i++) if (!strcmp (scode, slist[i].code) && !strcmp (ccode, slist[i].country)) return (i);
	return (-1);
}

int country_has_states (char *code, char *countries[], int n)
{
	int i;
	for (i = 0; i < n; i++) if (!strcmp (code, countries[i])) return (TRUE);
	return (FALSE);
}
