/*
 * $Id: extract_epic.c,v 1.21 2012/04/06 21:37:46 myself Exp $
 *
 * extract_epic.c reads the NEIC Epic global seismicity file and extracts a
 * subset of earthquakes based on the command-line selections.
 *
 * See file COPYING for Copyright information
 *
 * Geoware, (c) 2000-2012 P. Wessel
 */

#include "geoware.h"

/*
* This is the format of the data records:
0--------1---------2---------3---------4---------5---------6---------7---------8
CAT   YEAR   MO DA ORIG TIME   LAT     LON  DEP   MAGNITUDE  IEFM DTSVNWG DIST
PDE   1990   12 31 082637.37 -34.85  178.97 170 5.1 mb GS    .... .......      |
*/

#define EPIC_FILE "epic_quakes.dat"

struct EPIC_QUAKE {
	double lon, lat;
	float mag, z;
};

int q_compare_mz (const void *A, const void *B);
int q_compare_zm (const void *A, const void *B);
int q_compare_z (const void *A, const void *B);
int q_compare_m (const void *A, const void *B);

int main (int argc, char **argv) {
	int year, month, day, i, error = FALSE, verbose = FALSE, found = 0;
	int n_alloc, n_quakes = 0, first_year = 0, first_month = 0, first_day = 0, last_year = 10000, last_month = 12, last_day = 31;
	char line[BUFSIZ], file[BUFSIZ], *directory, *dir, *p, *sort_key = NULL;
	double low_z = 0.0, high_z = 10000.0, m_off = 0.0, m_scl = 1.0;
	double west = 0.0, east = 360.0, south = -90.0, north = 90.0, low_mag = 0.0, high_mag = 20.0;
	FILE *fp = NULL;
	struct EPIC_QUAKE *quakes;

	if ( (directory = getenv ("GMT_DATADIR")) == (char *)NULL) {
		fprintf (stderr, "%s: Environmental variable GMT_DATADIR not set!\n", argv[0]);
		exit (EXIT_FAILURE);
	}

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'R':
					sscanf (&argv[i][2], "%lf/%lf/%lf/%lf", &west, &east, &south, &north);
					break;
				case 'D':
					sscanf (&argv[i][2], "%d/%d/%d/%d/%d/%d", &first_year, &first_month, &first_day, &last_year, &last_month, &last_day);
					break;
				case 'M':
					sscanf (&argv[i][2], "%lf/%lf", &low_mag, &high_mag);
					break;
				case 'Z':
					sscanf (&argv[i][2], "%lf/%lf", &low_z, &high_z);
					break;
				case 'C':
					sscanf (&argv[i][2], "%lf/%lf", &m_off, &m_scl);
					break;
				case 'S':
					sort_key = &argv[i][2];
					break;
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
	}

	if (argc == 1 || error) {
		fprintf (stderr, "extract_epic %s - Extract earthquakes from NEIC EPIC data base\n\n", COMP_VERSION);
		fprintf (stderr, "usage: %s [-R<west/east/south/north>] [-C<m_off/m_scl>] [-S[z|m|zm|mz]\n", argv[0]);
		fprintf (stderr, "	[-M<low_mag/high_mag>] [-D<first_YYYY/MM/DD/last_YYYY/MM/DD>] [-V] [-Z<shallow/deep>]\n\n");
		fprintf (stderr, "	OPTIONS\n");
		fprintf (stderr, "	-C Offset magnitudes by <m_off> and multiply by <m_scl> to get symbols sizes [0/1]\n");
		fprintf (stderr, "	-D Set date range [all dates]\n");
		fprintf (stderr, "	-M Set lowest and highest magnitude [all magnitudes]\n");
		fprintf (stderr, "	-R Set region [0/360/-90/90]\n");
		fprintf (stderr, "	-S sort quakes for plotting purposes.  Append one or both of\n");
		fprintf (stderr, "	   z to sort quakes so deepest quakes are reported first\n");
		fprintf (stderr, "	   m to sort quakes so biggest quakes are reported first\n");
		fprintf (stderr, "	-V will report progress while working [Default is quiet]\n");
		fprintf (stderr, "	-Z Set depth range [all depths]\n");
		exit (EXIT_FAILURE);
	}
	
	
	n_alloc = CHUNK;
	quakes = (struct EPIC_QUAKE *) malloc ((size_t)(n_alloc * sizeof (struct EPIC_QUAKE)));
	
	dir = directory;
	while (!found && (p = strtok (dir, PATH_SEPARATOR))) {
		sprintf (file, "%s%c%s", p, DIR_DELIM, EPIC_FILE);
		if (!access (file, R_OK)) found = TRUE;
		dir = NULL;
	}
	if ((fp = fopen (file, "r")) == NULL) {
		fprintf (stderr, "%s:  Cannot open file %s!\n", argv[0], file);
		exit (EXIT_FAILURE);
	}
	else if (verbose)
		fprintf (stderr, "%s: Extract data from %s\n", argv[0], file);

	for (i = 0; i < 20; i++) fgets (line, BUFSIZ, fp);	/* Skip header */

	i = 0;
	while (fgets (line, BUFSIZ, fp)) {

		year = atoi (&line[6]);
		if (year < first_year || year > last_year) continue;
		if (line[13] == ' ' && line[14] == ' ') {	/* No month and day given */
			year = -1000;	/* Which causes mont/day checks to be bypassed below */
		}
		else {
			atoi (&line[13]);
			month = atoi (&line[13]);
			day = atoi (&line[16]);
		}
		if (year == first_year) {
			if (month < first_month) continue;
			if (month == first_month && day < first_day) continue;
		}
		if (year == last_year) {
			if (month > last_month) continue;
			if (month == last_month && day > last_day) continue;
		}
		quakes[i].lat = atof (&line[29]);
		if (quakes[i].lat < south || quakes[i].lat > north) continue;
		quakes[i].z = (float) atof (&line[44]);
		if (quakes[i].z < low_z || quakes[i].z > high_z) continue;
		quakes[i].mag = (float) atof (&line[48]);
		if (quakes[i].mag < low_mag || quakes[i].mag > high_mag) continue;
		quakes[i].lon = atof (&line[36]) - 360.0;
		while (quakes[i].lon < west) quakes[i].lon += 360.0;
		if (quakes[i].lon > east) continue;

		while (quakes[i].lon < 0.0) quakes[i].lon += 360.0;
		i++;
		if (i == n_alloc) {
			n_alloc += CHUNK;
			quakes = (struct EPIC_QUAKE *) realloc ((void *)quakes, (size_t)(n_alloc * sizeof (struct EPIC_QUAKE)));
		}
	}
	n_quakes = i;
	if (fp != stdin) fclose (fp);
	
	if (verbose) fprintf (stderr, "%s: Found %d earthquakes matching your search criteria\n", argv[0], n_quakes);
	if (sort_key && !strcmp (sort_key, "zm")) {
		if (verbose) fprintf (stderr, "%s: sorting quakes by depth, then magnitude...\n", argv[0]);
		qsort ((void *)quakes, (size_t) n_quakes, sizeof (struct EPIC_QUAKE), q_compare_zm);
	}
	else if (sort_key && !strcmp (sort_key, "mz")) {
		if (verbose) fprintf (stderr, "%s: sorting quakes by magnitude, then depth...\n", argv[0]);
		qsort ((void *)quakes, (size_t) n_quakes, sizeof (struct EPIC_QUAKE), q_compare_mz);
	}
	else if (sort_key && !strcmp (sort_key, "z")) {
		if (verbose) fprintf (stderr, "%s: sorting quakes by depth...\n", argv[0]);
		qsort ((void *)quakes, (size_t) n_quakes, sizeof (struct EPIC_QUAKE), q_compare_z);
	}
	else if (sort_key && !strcmp (sort_key, "m")) {
		if (verbose) fprintf (stderr, "%s: sorting quakes by magnitude...\n", argv[0]);
		qsort ((void *)quakes, (size_t) n_quakes, sizeof (struct EPIC_QUAKE), q_compare_m);
	}
	if (verbose) fprintf (stderr, "%s: Writing %d records to stdout...\n", argv[0], n_quakes);

	for (i = 0; i < n_quakes; i++)
		printf ("%g\t%g\t%g\t%g\n", quakes[i].lon, quakes[i].lat, quakes[i].z, (quakes[i].mag - m_off) * m_scl);
		
	free ((void *)quakes);
	if (verbose) fprintf (stderr, "%s: Done\n", argv[0]);
	
	exit (EXIT_SUCCESS);
}

int q_compare_zm (const void *A, const void *B) {
	struct EPIC_QUAKE *a, *b;
	a = (struct EPIC_QUAKE *)A;
	b = (struct EPIC_QUAKE *)B;
	if (a->z < b->z) return (1);
	if (a->z > b->z) return (-1);
	if (a->mag < b->mag) return (1);
	if (a->mag > b->mag) return (-1);
	return (0);
}

int q_compare_mz (const void *A, const void *B) {
	struct EPIC_QUAKE *a, *b;
	a = (struct EPIC_QUAKE *)A;
	b = (struct EPIC_QUAKE *)B;
	if (a->mag < b->mag) return (1);
	if (a->mag > b->mag) return (-1);
	if (a->z < b->z) return (1);
	if (a->z > b->z) return (-1);
	return (0);
}

int q_compare_z (const void *A, const void *B) {
	struct EPIC_QUAKE *a, *b;
	a = (struct EPIC_QUAKE *)A;
	b = (struct EPIC_QUAKE *)B;
	if (a->z < b->z) return (1);
	if (a->z > b->z) return (-1);
	return (0);
}

int q_compare_m (const void *A, const void *B) {
	struct EPIC_QUAKE *a, *b;
	a = (struct EPIC_QUAKE *)A;
	b = (struct EPIC_QUAKE *)B;
	if (a->mag < b->mag) return (1);
	if (a->mag > b->mag) return (-1);
	return (0);
}
