/*
 * $Id: extract_volcanoes.c,v 1.20 2012/04/06 21:37:46 myself Exp $
 *
 * extract_volcanoes.c reads the Smithsonian Holocene Volcanism file and extracts a
 * subset of volcanoes based on the command-line selections.
 *
 * See file COPYING for Copyright information
 *
 * Geoware, (c) 2000-2012 P. Wessel
 */

#include "geoware.h"

/*
* This is the format of the data records:
0--------1---------2---------3---------4---------5---------6---------7---------8---------9---------0---------1---------2---------3---------4
NUMBER    NAME                          LOCATION             LATITUDE LONGITUDE ELEV  TYPE                STATUS                  TIME FRAME
0100-01-  WEST EIFEL VOLC FIELD         Germany              50.17  N   6.85  E   600 Maars               Radiocarbon             D7
*/

#define VOLCFILE "volcanos_of_the_world.txt"

struct SMITHSONIAN {
	double lon, lat, z;
	char name[32];
};

int main (int argc, char **argv) {
	int i, j, error = FALSE, verbose = FALSE, header = FALSE;
	int n_alloc, n_volcanoes = 0, found = 0;
	char line[BUFSIZ], file[BUFSIZ], *directory, *dir, *p;
	double low_z = -20000.0, high_z = 20000.0;
	double west = 0.0, east = 360.0, south = -90.0, north = 90.0;
	FILE *fp = NULL;
	struct SMITHSONIAN *volcanoes;

	if ( (directory = getenv ("GMT_DATADIR")) == (char *)NULL) {
		fprintf (stderr, "%s: Environmental variable GMT_DATADIR not set!\n", argv[0]);
		exit (EXIT_FAILURE);
	}

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'R':
					sscanf (&argv[i][2], "%lf/%lf/%lf/%lf", &west, &east, &south, &north);
					break;
				case 'Z':
					sscanf (&argv[i][2], "%lf/%lf", &low_z, &high_z);
					break;
				case 'H':
					header = TRUE;
					break;
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
	}

	if (argc == 1 || error) {
		fprintf (stderr, "extract_volcanoes %s - Extract volcanoes from Smithsonian data base\n\n", COMP_VERSION);
		fprintf (stderr, "usage: extract_volcanoes [-H] [-R<west/east/south/north>] [-Z<zlow/zhigh>]\n\n");
		fprintf (stderr, "	OPTIONS\n");
		fprintf (stderr, "	-H Output a 1-record header [no header]\n");
		fprintf (stderr, "	-R Set region [0/360/-90/90]\n");
		fprintf (stderr, "	-V will report progress while working [Default is quiet]\n");
		fprintf (stderr, "	-Z Set elevation range [all elevations]\n");
		exit (EXIT_FAILURE);
	}
	
	dir = directory;
	while (!found && (p = strtok (dir, PATH_SEPARATOR))) {
		sprintf (file, "%s%c%s", p, DIR_DELIM, VOLCFILE);
		if (!access (file, R_OK)) found = TRUE;
		dir = NULL;
	}
	if ((fp = fopen (file, "r")) == NULL) {
		fprintf (stderr, "%s:  Cannot open file %s!\n", argv[0], file);
		exit (EXIT_FAILURE);
	}
	else if (verbose)
		fprintf (stderr, "%s: Extract data from %s\n", argv[0], file);
	
	n_alloc = CHUNK;
	volcanoes = (struct SMITHSONIAN *) malloc ((size_t)(n_alloc * sizeof (struct SMITHSONIAN)));
	
	for (i = 0; i < 2; i++) fgets (line, BUFSIZ, fp);	/* Skip header */

	i = 0;
	while (fgets (line, BUFSIZ, fp)) {

		volcanoes[i].lat = atof (&line[61]);
		if (line[68] == 'S') volcanoes[i].lat = -volcanoes[i].lat;
		if (volcanoes[i].lat < south || volcanoes[i].lat > north) continue;
		volcanoes[i].z = (isdigit (line[84])) ? atof (&line[80]) : 0.0;
		if (volcanoes[i].z < low_z || volcanoes[i].z > high_z) continue;
		volcanoes[i].lon = atof (&line[70]);
		if (line[78] == 'W') volcanoes[i].lon = -volcanoes[i].lon;
		volcanoes[i].lon -= 360.0;
		while (volcanoes[i].lon < west) volcanoes[i].lon += 360.0;
		if (volcanoes[i].lon > east) continue;
		strncpy (volcanoes[i].name, &line[10], 30);
		j = 29;
		while (j > 0 && volcanoes[i].name[j] == ' ') j--;
		volcanoes[i].name[j+1] = '\0';
		while (volcanoes[i].lon < 0.0) volcanoes[i].lon += 360.0;
		i++;
		if (i == n_alloc) {
			n_alloc += CHUNK;
			volcanoes = (struct SMITHSONIAN *) realloc ((void *)volcanoes, (size_t)(n_alloc * sizeof (struct SMITHSONIAN)));
		}
	}
	n_volcanoes = i;
	fclose (fp);
	
	if (verbose) fprintf (stderr, "%s: Found %d volcanoes matching your search criteria\n", argv[0], n_volcanoes);
	
	if (n_volcanoes) {
		if (header) printf ("#lon\tlat\televation\tname\n");
		for (i = 0; i < n_volcanoes; i++)
			printf ("%g\t%g\t%g\t%s\n", volcanoes[i].lon, volcanoes[i].lat, volcanoes[i].z, volcanoes[i].name);
	}
		
	free ((void *)volcanoes);
	if (verbose) fprintf (stderr, "%s: Done\n", argv[0]);
	
	exit (EXIT_SUCCESS);
}
