/*
 *	$Id: extrdata.c,v 1.87 2012/04/06 21:37:46 myself Exp $
 *
 * extrdata is used on both the Windows and Unix version of the GMT
 * Companion CD/DVD-ROMs to determine which data files are to be installed.
 * It does so by asking questions and then starts extracting and
 * decoding data using system calls (WIN32) or shell commands (UNIX).
 * We assume that the installation has gone successful to this point
 * so that the <INSTALLDIR>/DATA/{misc,grids,img} directories exist.
 * The Unix version actually writes all the commands to stdout and
 * the master install.sh script will save those to a script file and
 * then execute it as a batch job.  The logic assumes that extrdata
 * is run from the CD/DVD-ROM location.
 *
 * 5-26-01:	Add -EB to geodecoder for *.img files since they should
 *		always be written as Bigendian
 * 10-13-01:	Add option to pass L or U for lower or upper case file
 *		names [Default makes no change]
 * 12-07-01:	Now uses new versions of geodecoder/geobzip2 that has
 *		built-in bzip2lib support.  We no longer need the dat
 *		files to list the bzip2 file sizes.
 * 03-10-02:	Dynamic allocation of structures.
 * 03-29-02:	Now use grdzip and new syntax for geobzip2
 * 01-15-06:	Make changes to support category with files in subdirs
 * 02-05-06:	Make changes to support subsets of categories (SRTM)
 *		  as well as option to just copy SRTM files.
 * 04-10-09:	Can handle TTT's T category and support for UNIX subdir
 *
 * Geoware, (c) 2000-2012 P. Wessel
 */

#include "geoware.h"

/* Change these parameters if you add much more data */

#define N_MEDIA			6
#define MAX_CATEGORIES		20
#define DATA_CHUNK		50
#define KB			1024.0
#define MB			(KB * KB)
#define GB			(MB * KB)

#ifdef WIN32
#define COPY	"copy"
#else
#define COPY	"cp -f"
#endif
char *sdir[3] = {
	"misc",
	"grids",
	"img"
};

char *unix2dos = " -u ";
char *none = " ";

struct DATA {
	char fullname[128];	/* Full name of data file */
	char dosname[16];	/* 8.3 DOS name */
	char synopsis[128];	/* Description of data */
	int type;		/* What kind of data */
	double full_size;	/* Size of expanded file in bytes */
	int install;		/* TRUE/FALSE to install or not */
	int ascii;		/* TRUE is file is (Unix) ascii and need CR/LF translation */
	int decode;		/* 2 = grdzip -d (.img), 1 = grdzip -d (.i2), 0 = no decode */
} *data;

struct CATEGORY {
	char name[8];		/* Short name of category */
	char description[80];	/* Descriptive title */
	char dir[80];		/* Subdir to use for all-or-nothing data */
	int all_or_nothing;	/* FALSE is we should allow questions for individual files */
	int copy;		/* TRUE if we should ask if plain copying is desired instead of install */
	int special;		/* For special treatment */
	int len;		/* Number of letters in name to use when matching text strings */
	double total;		/* Size of files in this category */
} category[MAX_CATEGORIES];
	
int find_category (char *key);
int find_dataset (char *key);
void str_tolower (char *value);
void str_toupper (char *value);

int n_categories = 0, n_data = 0;

int main (int argc, char **argv) {

	int i, j, id, n_words, done = 0, media_no, this_srtm = -1, is_srtm = -1, look_for_next = 0;
	int SWBD = -1, n_step_up = 0, down = 0;
	size_t n_alloc = 0;
	char unit, line[32], record[BUFSIZ], name[BUFSIZ], title[BUFSIZ], *dir = NULL, CDROM[256], cmd[BUFSIZ], rev[16];
	char *a_or_b, w1[64], w2[64], w3[64], w4[64], w5[64], w6[64], w7[64], kase = '-', *pdir = NULL;
	char GEOBZIP2[256], GRDZIP[256], file[256];
	char *prefix_lower[N_MEDIA] = {"gmtcomp", "gmtcompa", "gmtcompb", "gmtcompc", "tttcdrom", "sdkcdrom"};
	char *prefix_upper[N_MEDIA] = {"GMTCOMP", "GMTCOMPA", "GMTCOMPB", "GMTCOMPC", "TTTCDROM", "SDKCDROM"};
	char *media[N_MEDIA] = {"DVD-ROM", "DVD-ROM (A)", "DVD-ROM (B)", "DVD-ROM (C)", "CD-ROM", "CD-ROM"};
	char fullname[256];	/* Full name of data file */
	double size, total_size = 0.0;
	FILE *fp;
	
#ifdef WIN32
	int CD, drv, k;
	
	/* Search for the proper drive to install from */
	
	for (drv = 'D', CD = 0; CD == 0 && drv <= 'Z'; drv++) {
#ifdef DEBUG
		fprintf (stderr, "%s: Examine drive %c.  ", argv[0], drv);
#endif
		sprintf (name, "%c:%cGEOWARE.TXT", drv, DIR_DELIM);
		if (!access (name, R_OK)) {	/* OK, this drive has a Geoware.txt file. Is it the Geoware CD? */
#ifdef DEBUG
			fprintf (stderr, "Found %s  ", name);
#endif
			sprintf (name, "%c:%cINSTALL.SH", drv, DIR_DELIM);
			if (!access (name, R_OK)) {
				CD = drv;	/* OK, found the Geoware CD in top dir */
#ifdef DEBUG
				fprintf (stderr, "Found %s  ", name);
#endif
			}
			sprintf (name, "%c:%cUNIX%cINSTALL.SH", drv, DIR_DELIM, DIR_DELIM);
			if (!access (name, R_OK)) {
				CD = drv;	/* OK, found the Geoware CD in a UNIX subdir */
				down = 1;
#ifdef DEBUG
				fprintf (stderr, "Found %s  ", name);
#endif
			}
		}
#ifdef DEBUG
		fprintf (stderr, "\n");
#endif
	}
	if (CD == 0) {
		fprintf (stderr, "%s: No Geoware media currently mounted!\n", argv[0]);
		exit (EXIT_FAILURE);
	}
		
	/* Cd to the correct drive */
	
	sprintf (name, "%c:%c", CD, DIR_DELIM);
	chdir (name);
#endif

	(void *) getcwd (CDROM, 256);
#ifdef DEBUG
	fprintf (stderr, "%s: Installing from %s\n", argv[0], CDROM);
#endif

	if (argc == 1) {
		fprintf (stderr, "extrdata %s - The Geoware Decompression Utility\n\n", UTILS_VERSION);
		fprintf (stderr, "usage: extrdata <INSTALLDIR> [-C<case>] [-D<n>] [-P<bindir>]\n\n");
		fprintf (stderr, "The Geoware Decompression Utility extrdata is used to extract\n");
		fprintf (stderr, "densily compressed and packed data files from the Geoware GMT\n");
		fprintf (stderr, "Companion CD/DVD-ROMs.\n");
		fprintf (stderr, "\t-C Specify <case> as L|l (Lower) or U|u (upper) [No case change].\n");
		fprintf (stderr, "\t-D Specify number of dir nesting to undo [0].\n");
		fprintf (stderr, "\t-P Specify path to grdzip/geobzip2 executables [Rely on system PATH].\n");
#ifdef WIN32
		fprintf (stderr, "\n==> Hit <ENTER> to continue: ");
		fgets (line, 32, stdin);
#endif
		exit (EXIT_FAILURE);
	}
	
	kase = '-';	/* Default is to leave file names as they are */
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {	/* Options were passed */
			switch (argv[i][1]) {
				case 'C':	/* Specify upper or lower case file names */
					if (argv[i][2] == 'l' || argv[i][2] == 'L') kase = 'L';
					if (argv[i][2] == 'u' || argv[i][2] == 'U') kase = 'U';
					break;
				case 'D':	/* Must issue "cd .." this many times */
					n_step_up = atoi (&argv[i][2]);
					break;
				case 'P':	/* Bin directory for grdzip, geobzip2 executables */
					pdir = &argv[i][2];
					break;
				default:
					fprintf (stderr, "%s: Bad option passed (%s)\n", argv[0], argv[i]);
					break;
			}
		}
		else
			dir = argv[1];	/* Installation directory */
	}

	/* Look for a gmtcomp?.dat file in both upper/lower case in current dir (CD) */
	
	for (i = 0, media_no = -1; media_no == -1 && i < N_MEDIA; i++) {
		(down) ? sprintf (name, "unix%c%s.dat", DIR_DELIM, prefix_lower[i]) : sprintf (name, "%s.dat", prefix_lower[i]);
		if ((fp = fopen (name, "r"))) {
			media_no = i;
#ifdef DEBUG
			fprintf (stderr, "%s: Found %s\n", argv[0], name);
#endif
		}
		else {
			(down) ? sprintf (name, "UNIX%c%s.DAT", DIR_DELIM, prefix_upper[i]) : sprintf (name, "%s.DAT", prefix_upper[i]);
			if ((fp = fopen (name, "r"))) {
				media_no = i;
#ifdef DEBUG
				fprintf (stderr, "%s: Found %s\n", argv[0], name);
#endif
			}
		}
	}

	if (media_no == -1) {
		fprintf (stderr, "%s: Cannot find/open a Geoware *.dat file\n\n", argv[0]);
#ifdef WIN32
		fprintf (stderr, "\n==> Hit <ENTER> to continue: ");
		fgets (line, 32, stdin);
#endif
		exit (EXIT_FAILURE);
	}

	if (pdir) {	/* Use specified bin directory */
		sprintf (GEOBZIP2, "%s%cgeobzip2 -V", pdir, DIR_DELIM);
		sprintf (GRDZIP, "%s%cgrdzip -de -v", pdir, DIR_DELIM);
	}
	else {	/* Default to INSTALLDIR/bin for Win32 and use system PATH otherwise */
#ifdef WIN32
		sprintf (GEOBZIP2, "%s%cbin%cgeobzip2 -V", dir, DIR_DELIM, DIR_DELIM);
		sprintf (GRDZIP, "%s%cbin%cgrdzip -de -v", dir, DIR_DELIM, DIR_DELIM);
#else
		sprintf (GEOBZIP2, "geobzip2 -V");
		sprintf (GRDZIP, "grdzip -de -v");
#endif
	}
	
	/* Allocate initial memory for data structure */
	
	n_alloc = DATA_CHUNK;
	data = (struct DATA *) calloc (n_alloc, sizeof (struct DATA));
	memset ((void *)category, 0, (size_t)(sizeof (struct CATEGORY) * MAX_CATEGORIES));
	
	while (fgets (record, BUFSIZ, fp)) {
		switch (record[0]) {
			case '#':		/* Skip comment lines */
				if (look_for_next) this_srtm++;
				continue;
				break;
				
			case '+':		/* Category entry */
				if (n_categories == MAX_CATEGORIES) {
					fprintf (stderr, "%s: Exceeding max number of categories (%d)\n", argv[0], MAX_CATEGORIES);
					exit (EXIT_FAILURE);
				}
				i = 2;
				if (record[1] == '!') {		/* ! means either install entire category or nothing */
					category[n_categories].all_or_nothing = TRUE;
					i = 3;
				}
				else if (record[1] == '+') {		/* + = ! plus option to copy rather than install */
					category[n_categories].all_or_nothing = TRUE;
					category[n_categories].copy = 1;
					i = 3;
				}
				else if (record[1] == '-') {		/* - = + except we only copy the files */
					category[n_categories].all_or_nothing = TRUE;
					category[n_categories].copy = 2;
					i = 3;
				}
				sscanf (&record[i], "%s %[^\n]", name, title);
				strcpy (category[n_categories].description, title);
				strcpy (category[n_categories].name, name);
				if (!strncmp (name, "SRTM", 4)) {	/* SRTM is special */
					category[n_categories].special = 1;
					category[n_categories].len = 4;		/* So SRTM-XX is seen as SRTM */
					if (this_srtm == -1) this_srtm = is_srtm = n_categories;
				}
				else
					category[n_categories].len = strlen (name);
				strncpy (category[n_categories].dir, name, category[n_categories].len);
#ifdef WIN32
				for (k = 0; k < (int)strlen (category[n_categories].dir); k++) if (category[n_categories].dir[k] == '/') category[n_categories].dir[k] = DIR_DELIM;
#endif
				n_categories++;
				break;
				
			case '-':		/* Individual data set description */
				sscanf (&record[2], "%s %[^\n]", name, title);
				id = find_dataset (name);
				strcpy (data[id].synopsis, title);
				break;
				
			default:		/* File info records [we skip if n_words == 5 which are program records] */
				n_words = sscanf (record, "%s %s %s %s %s %s %s", w1, w2, w3, w4, w5, w6, w7);
				if (n_words == 7) {	/* Data record */
					strcpy (data[n_data].fullname, w1);
					strcpy (data[n_data].dosname, w2);
					data[n_data].ascii = (w7[0] == 'A' || w7[0] == 'a');
					if (w7[0] == 'P') {	/* Must be GMT coastline files */
						data[n_data].full_size = atoi (w5);
						data[n_data].type = find_category (w4);
						data[n_data].decode = 3;
					}
					else if (w7[0] == 'T') {	/* TTT *.i2 files */
						data[n_data].full_size = atoi (w4);
						data[n_data].type = find_category (w5);
						data[n_data].decode = 4;
					}
					else {			/* Must be D, I, or G */
						data[n_data].full_size = atoi (w4);
						data[n_data].type = find_category (w5);
						data[n_data].decode = (w7[0] == 'D') ? 0 : ((w7[0] == 'G') ? 1 : 2);
					}
					if (data[n_data].type == is_srtm) {
						look_for_next = 1;
						data[n_data].type = this_srtm;
					}
					if (data[n_data].full_size == 0.0) data[n_data].decode = 9;	/* Directory to create */
					n_data++;
					if (n_data == (int)n_alloc) {	/* Get more memory */
						n_alloc += DATA_CHUNK;
						data = (struct DATA *) realloc ((void *)data, (size_t) (n_alloc * sizeof (struct DATA)));
					}
				}
				else if (n_words != 5) {
					fprintf (stderr, "%s: Malformed record? (%s)\n", argv[0], record);
					exit (EXIT_FAILURE);
				}
				break;
		}
	}
	fclose (fp);
				
#ifdef DEBUG
	fprintf (stderr, "CAT: %d   DATA: %d\n\n", n_categories, n_data);
	for (i = 0; i < n_data; i++) fprintf (stderr, "%-30s%-15s%10u  %-8s%4d\n", data[i].fullname, data[i].dosname, (unsigned int)data[i].full_size, category[data[i].type].name, data[i].decode);
#endif
				
	for (i = 0; i < n_data; i++) category[data[i].type].total += data[i].full_size;
	for (j = 0; j < n_categories; j++) total_size += category[j].total;
	size = total_size / GB;

	sscanf ("$Revision: 1.87 $", "%*s %s", rev);
	fprintf (stderr, "\n>>> Geoware GMT Companion CD/DVD-ROM Data Decompression Utility  v. %s <<<\n\n", rev);
	fprintf (stderr, "About %.1f Gigabyte of data may be extracted from this %s.\n", size, media[media_no]);
	fprintf (stderr, "Answer questions below to select which data you intend to use.\n");
	fprintf (stderr, "The data sets are:\n\n");
	
	if (media_no == 1 || media_no == 2) SWBD = find_category ("SWBD");	/* Vols. A + B */
	
	for (j = 0; j < n_categories; j++) {
		fprintf (stderr, "%2d. %7s : %s\n", j+1, category[j].name, category[j].description);
	}
	while (!done) {
		total_size = 0.0;
		fprintf (stderr, "\n==> Install (a)ll, (s)ome, or (n)one of the data sets (a/s/n) [s]: ");
		fgets (line, 32, stdin);
		if (line[0] == 'N' || line[0] == 'n') {				/* Dont want any kind of data, quit */
			fprintf (stderr, "%s: No data sets selected or installed - exiting.\n", argv[0]);
			exit (EXIT_SUCCESS);
		}
		else if (line[0] == 'A' || line[0] == 'a') {	/* Want everything, easy! */
			for (j = 0; j < n_categories; j++) {
				for (i = 0; i < n_data; i++) if (data[i].type == j) data[i].install = 1;
				total_size += category[j].total;
				if (category[j].copy == 1) category[j].copy = FALSE;	/* No longer an option to copy since we want to install all */
			}
		}
		else {								/* Must ask questions for each data category */
			fprintf (stderr, "\n\t>>> Choose the data categories of interest <<<\n");
			for (j = 0; j < n_categories; j++) {
				size = category[j].total / KB;
				unit = 'K';
				if (size >= KB) {
					size /= KB;
					unit = 'M';
					if (size >= KB) {
						size /= KB;
						unit = 'G';
					}
				}
				fprintf (stderr, "\n    %d. %s data (total of %.1f %cb).\n", j+1, category[j].description, size, unit);
				if (category[j].all_or_nothing)
					fprintf (stderr, "==> Install (a)ll or (n)one of these data (a/n) [a]: ");
				else
					fprintf (stderr, "==> Install (a)ll, (s)ome, or (n)one of these data (a/s/n) [a]: ");
				fgets (line, 32, stdin);
				if (category[j].all_or_nothing && (line[0] == 'S' || line[0] == 's')) line[0] = 'A';	/* Force some to become all */
				if (line[0] == 'N' || line[0] == 'n') {				/* Not interested in these data */
					for (i = 0; i < n_data; i++) if (data[i].type == j) data[i].install = 0;
				}
				else if (line[0] == 'A' || line[0] == 'a' || line[0] == '\0' || line[0] == '\r' || line[0] == '\n') {	/* Want everything, easy! */
					for (i = 0; i < n_data; i++) if (data[i].type == j) data[i].install = 1;
					total_size += category[j].total;
					if (category[j].copy == 1) {
						fprintf (stderr, "\tDue to large size, category can either be installed or just copied.\n");
						fprintf (stderr, "\tCopying reduces the required disk space considerably but\n");
						fprintf (stderr, "\tmeans un-grdzipping will take place when data are requested.\n");
						fprintf (stderr, "\t==> Shall we (i)nstall or (c)opy these data (i/c) [c]: ");
						fgets (line, 32, stdin);
						if (line[0] == 'I' || line[0] == 'i') category[j].copy = FALSE;
					}
				}
				else {								/* Must ask question for each data set */
					for (i = 0; i < n_data; i++) {
						if (data[i].type != j) continue;
						size = (data[i].full_size / KB);
						unit = 'K';
						if (size >= KB) {
							size /= KB;
							unit = 'M';
							if (size >= KB) {
								size /= KB;
								unit = 'G';
							}
						}
						fprintf (stderr, "--> Install %s (%.1f %cb (y/n) [y]: ", data[i].synopsis, size, unit);
						fgets (line, 32, stdin);
						if (line[0] == 'Y' || line[0] == 'y' || line[0] == '\0' || line[0] == '\r' || line[0] == '\n') {
							data[i].install = 1;
							total_size += (double)data[i].full_size;
						}
					}
				}
			}
		}
		total_size /= MB;
		unit = 'M';
		if (total_size >= KB) {
			total_size /= KB;
			unit = 'G';
		}
		fprintf (stderr, "\n==> OK to install %.1f %cb of data (y/n) [y]: ", total_size, unit);
		fgets (line, 32, stdin);
		if (line[0] == 'Y' || line[0] == 'y' || line[0] == '\0' || line[0] == '\r' || line[0] == '\n') done = 1;
	}
	
	
	/* List what to install */

	fprintf (stderr, "\nDecompress and decode selected data sets:\n\n");

#ifndef WIN32
	printf ("#!/bin/sh\n");
	printf ("# Geoware Companion %s data extraction/install script\n", media[media_no]);
	printf ("# Generating command: extrdata %s\n#\n", dir);
	printf ("cd %s\n", CDROM);
	for (i = 0; i < n_step_up; i++) printf ("cd ..\n");
#endif
	for (i = 0; i < n_data; i++) {
		if (data[i].decode == 9) {	/* We always create the directory tree */
			sprintf (file, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[category[data[i].type].special], DIR_DELIM, data[i].fullname);
#ifdef WIN32
			/* Change any directory slashes in final file names (i.e., for DCW, WOA01) to DOS backward slashes */
			for (k = 0; k < (int)strlen(file); k++) if (file[k] == '/') file[k] = DIR_DELIM;
			sprintf (cmd, "mkdir %s", file);
			system (cmd);
#else
			sprintf (cmd, "mkdir -p %s", file);
			printf ("%s\n", cmd);
#endif
			continue;
		}
		if (data[i].install == 0) continue;
#ifdef WIN32
		/* Change any directory slashes in final file names (i.e., for DCW, WOA01) to DOS backward slashes */
		for (k = 0; k < (int)strlen (data[i].fullname); k++) if (data[i].fullname[k] == '/') data[i].fullname[k] = DIR_DELIM;
#endif
		a_or_b = (data[i].ascii) ? unix2dos : none;
		if (data[i].decode == 2) {	/* *.img files, save in Bigendian format */
			sprintf (file, "DATA%c%s", DIR_DELIM, data[i].dosname);
			if (kase == 'L') str_tolower (file);
			if (kase == 'U') str_toupper (file);
			sprintf (cmd, "%s %s -EB -F%s%cDATA%c%s%c%s", GRDZIP, file, dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
		}
		else if (data[i].decode == 1) {	/* *.i2 Grid files */
			if (category[data[i].type].special) {
				sprintf (fullname, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
				sprintf (file, "DATA%c%s%c%s", DIR_DELIM, category[data[i].type].dir, DIR_DELIM, data[i].dosname);
			}
			else {
				sprintf (fullname, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
				sprintf (file, "DATA%c%s", DIR_DELIM, data[i].dosname);
			}
			if (kase == 'L') str_tolower (file);
			if (kase == 'U') str_toupper (file);
			if (category[data[i].type].all_or_nothing) {	/* Install under subdir */
				if (category[data[i].type].copy) {	/* Just copy, dont expand. Replace trailing .i2 with .gbz */
					sprintf (cmd, "%s %s %s", COPY, file, fullname);
					j = strlen (cmd) - 2;
					cmd[j++] = 'g';	cmd[j++] = 'b';	cmd[j++] = 'z';cmd[j] = '\0';
#ifdef WIN32
					fprintf (stderr, "%s\n", cmd);
#else
					printf ("echo %s\n", cmd);
#endif
				}
				else
					sprintf (cmd, "%s %s -F%s", GRDZIP, file, fullname);
			}
			else
				sprintf (cmd, "%s %s -F%s", GRDZIP, file, fullname);
		}
		else if (data[i].decode == 4) {	/* TTT *.i2 Grid files */
			sprintf (fullname, "%s%cdata%c%s", dir, DIR_DELIM, DIR_DELIM, data[i].fullname);
			sprintf (file, "DATA%c%s", DIR_DELIM, data[i].dosname);
			if (kase == 'L') str_tolower (file);
			if (kase == 'U') str_toupper (file);
			sprintf (cmd, "%s %s -F%s", GRDZIP, file, fullname);
		}
		else {
			if (!strcmp (category[data[i].type].name, "WOA01")) {
				sprintf (file, "DATA%cWOA01%c%s", DIR_DELIM, DIR_DELIM, data[i].dosname);
				sprintf (fullname, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
			}
			else if (!strcmp (category[data[i].type].name, "DCW")) {
				sprintf (file, "DATA%cDCW%c%s", DIR_DELIM, DIR_DELIM, data[i].dosname);
				sprintf (fullname, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
			}
			else if (data[i].type == SWBD) {
				sprintf (fullname, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
				sprintf (file, "DATA%cSWBD%c%s", DIR_DELIM, DIR_DELIM, data[i].dosname);
			}
			else {
				sprintf (file, "DATA%c%s", DIR_DELIM, data[i].dosname);
				sprintf (fullname, "%s%cDATA%c%s%c%s", dir, DIR_DELIM, DIR_DELIM, sdir[data[i].decode], DIR_DELIM, data[i].fullname);
			}
			if (kase == 'L') str_tolower (file);
			if (kase == 'U') str_toupper (file);
			if (category[data[i].type].copy) { 	/* Just copy, dont expand. */
				sprintf (cmd, "%s %s %s", COPY, file, fullname);
#ifdef WIN32
				fprintf (stderr, "%s\n", cmd);
#else
				printf ("echo %s\n", cmd);
#endif
			}
			else
				sprintf (cmd, "%s %s -I%u%s-F%s", GEOBZIP2, file, (unsigned int)data[i].full_size, a_or_b, fullname);
		}
#ifdef WIN32
		system (cmd);
#else
		printf ("%s\n", cmd);
#endif
	}

#ifdef WIN32
	fprintf (stderr, "%s: Installed %.2lf %cb of data\n", argv[0], total_size, unit);
	fprintf (stderr, "\n==> Hit <ENTER> to continue: ");
	fgets (line, 32, stdin);
#else
	printf ("echo \"%s: Installed %.2f %cb of data\"\n", argv[0], total_size, unit);
	printf ("cd %s\n", dir);
#endif
	free ((void *)data);

	exit (EXIT_SUCCESS);
}

int find_category (char *key)
{	/* Given string with category name, return category index */
	int j, k;
	
	for (k = 0, j = -1; j == -1 && k < n_categories; k++) if (!strncmp (key, category[k].name, category[k].len)) j = k;
	if (j == -1) {
		fprintf (stderr, "Error in find_category using key %s\n", key);
		exit (EXIT_FAILURE);
	}
	return (j);
}

int find_dataset (char *key)
{	/* Given string with dataset name, return data set index */
	int j, k;
	
	for (k = 0, j = -1; j == -1 && k < n_data; k++) if (!strcmp (key, data[k].dosname)) j = k;
	if (j == -1) {
		fprintf (stderr, "Error in find_dataset using key %s\n", key);
		exit (EXIT_FAILURE);
	}
	return (j);
}

void str_tolower (char *value)
{
	/* Convert entire string to lower case */
	int i, c;
	for (i = 0; value[i]; i++) {
		c = (int)value[i];
		value[i] = (char) tolower (c);
	}
}

void str_toupper (char *value)
{
	/* Convert entire string to upper case */
	int i, c;
	for (i = 0; value[i]; i++) {
		c = (int)value[i];
		value[i] = (char) toupper (c);
	}
}
