/*	$Id: geobzip2.c,v 1.24 2012/04/06 21:37:46 myself Exp $
 *
 * geobzip2 does a bzip2 -d and displays a progress bar if final size is known.
 *
 * Geoware, (c) 2000-2012 P. Wessel
 */

#include "geoware.h"
#include "bzlib.h"

int main (int argc, char **argv) {
	int i, k, m_len, width, n, n_out = 0, out_sign, bzerror, comp_size = 0, final_size = 0, unit = 1;
	int dos2unix = 0, unix2dos = 0, error = 0, verbose = 0;
	char out_marker[51] = "----1----2----3----4----5----6----7----8----9----|";
	char *infile = NULL, *outfile = NULL, buffer[BUFSIZ], CR = '\r', *u[3] = {"Gb", "Mb", "kb"};
	char message[BUFSIZ];
	double in_Mb, out_Mb, pack, two_percent, last_out, b_to_Mb;
	FILE *fpi, *fpo;
	BZFILE *bz;
	struct stat buf;
	
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'd':	/* Convert DOS ascii file to Unix */
				case 'D':
					dos2unix = 1;
					break;
				case 'u':	/* Convert Unix ascii file to DOS */
				case 'U':
					unix2dos = 1;
					break;
				case 'I':	/* Compressed file size (when reading stdin) */
				case 'i':
					final_size = atoi (&argv[i][2]);
					break;
				case 'F':	/* Output file */
				case 'f':
					outfile = &argv[i][2];
					break;
				case 'v':	/* Verbose */
				case 'V':
					verbose = 1;
					break;
				default:
					error++;
					break;
			}
		}
		else
			infile = argv[i];
	}
	
	if (argc == 0 || error) {
		fprintf (stderr, "geobzip2 %s - counting of bzip2 extraction\n\n", UTILS_VERSION);
		fprintf (stderr, "usage: geobzip2 [compressedfile] [-D|U] [-F<decompressedfile>] [-I<finalsize>] [-V]\n");
		fprintf (stderr, "\t-D Convert DOS ascii file to Unix\n");
		fprintf (stderr, "\t-F Output file name [Default writes to stdout]\n");
		fprintf (stderr, "\t-I Give unpacked file size (for -V).\n");
		fprintf (stderr, "\t-U Convert Unix ascii file to DOS\n");
		fprintf (stderr, "\t-V Verbose\n");
		exit (EXIT_FAILURE);
	}

	memset ((void *)message, 0, BUFSIZ);
	if (infile) {
		if ((fpi = fopen (infile, "rb")) == NULL) {
			fprintf (stderr, "geobzip2: Cannot open/find file %s\n", infile);
			exit (EXIT_FAILURE);
		}
		sprintf (message, "--> %s to ", infile);
		stat (infile, &buf);	
		comp_size = buf.st_size;
	}
	else {
		fpi = stdin;
#if defined(_WIN32) || defined(WIN32)
		_setmode (fileno (stdin), _O_BINARY);
#elif defined(__EMX__)
		_fsetmode (stdin, "b");
#endif		
		sprintf (message, "--> <stdin> to ");
	}
	if (verbose) fprintf (stderr, "%s", message);
	m_len = strlen (message);
	memset ((void *)message, 0, BUFSIZ);
	if (outfile) {
		int start, skip;
		if ((fpo = fopen (outfile, "wb")) == NULL) {
			fprintf (stderr, "geobzip2: Cannot create file %s\n", outfile);
			exit (EXIT_FAILURE);
		}
		sprintf (message, "%s", outfile);
		width = 80 - m_len;
		if ((m_len = strlen (message)) > width) {	/* Must shorten message by using ellipses in the middle */
			skip = m_len - width + 3;
			i = start = m_len/2 - skip/2;
			for (k = 0; k < 3; k++, i++) message[i] = '.';	/* Ellipses ... */
			k = start + skip;
			while (message[k]) message[i++] = message[k++];
			message[i] = '\0';
		}
	}
	else {
		fpo = stdout;
#if defined(_WIN32) || defined(WIN32)
		_setmode (fileno (stdout), _O_BINARY);
#elif defined(__EMX__)
		_fsetmode (stdout, "b");
#endif		
		strcpy (message, "<stdout>");
	}
	if (verbose) fprintf (stderr, "%s\n", message);

	bz = BZ2_bzReadOpen (&bzerror, fpi, 0, 0, NULL, 0);
	if (bzerror != BZ_OK) {
		fprintf (stderr, "geobzip2: Error initializing compressed input\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	
	if (verbose && final_size <= 0) {
		fprintf (stderr, "geobzip2: finalsize not known - no progress-bar possible\n");
		verbose = 0;
	}

	b_to_Mb = 1.0 / (1024.0 * 1024.0);
	two_percent = 0.02 * final_size;
	out_sign = 0;
	last_out = 0.0;

	if (verbose) fprintf (stderr, "|");

	while ((n = BZ2_bzRead (&bzerror, bz, (void *)buffer, BUFSIZ)) > 0) {
		if (unix2dos) {	/* Must do char by char check */
			for (i = 0; i < n; i++) {
				if (buffer[i] == '\n') fwrite ((void *)&CR, sizeof(char), 1, fpo);
				fwrite ((void *)&buffer[i], sizeof(char), 1, fpo);
			}
		}
		else if (dos2unix) {	/* Must do char by char check */
			for (i = 0; i < n; i++) {
				if (buffer[i] != '\r') fwrite ((void *)&buffer[i], sizeof(char), 1, fpo);
			}
		}
		else
			fwrite ((void *)buffer, sizeof(char), n, fpo);
		n_out += n;
		if (verbose) {
			while (n_out > (int) (last_out)) {
				fprintf (stderr, "%c", out_marker[out_sign]);
				last_out += two_percent;
				out_sign++;
			}
		}
	}
	BZ2_bzReadClose (&bzerror, bz);
	if (bzerror != BZ_OK) {
		fprintf (stderr, "geobzip2: Error closing compressed file\n");
		exit (EXIT_FAILURE);
	}
	if (fpi != stdin) fclose (fpi);
	if (fpo != stdout) fclose (fpo);
	
	if (verbose) {
		while (n_out > (int) (last_out + 0.5) && out_sign < 50) {
			fprintf (stderr, "%c", out_marker[out_sign]);
			last_out += two_percent;
			out_sign++;
		}
	}
	if (final_size && n_out != final_size) {
		fprintf (stderr, "geobzip2: Warning: given finalsize (%d) does not match actual size (%d)!\n", final_size, n_out);
	}
	out_Mb = n_out * b_to_Mb;
	in_Mb  = comp_size * b_to_Mb;
	if (out_Mb < 1.0) {	/* Switch to kb */
		out_Mb *= 1024.0;
		in_Mb *= 1024.0;
		unit = 2;
	}
	else if (out_Mb > 1024.0) {	/* Switch to Gb */
		out_Mb /= 1024.0;
		in_Mb /= 1024.0;
		unit = 0;
	}
	pack = 100.0 * in_Mb / out_Mb;
	if (verbose) fprintf (stderr, "%6.1f -> %6.1f %s [%4.1f %%]\n", in_Mb, out_Mb, u[unit], pack);

	exit (EXIT_SUCCESS);
}
