/*	$Id: grdzip.c,v 1.42 2012/04/06 21:37:46 myself Exp $
 *
 * This software is described in more detail in the article
 *
 * "Compression of Large Data Grids for Internet Transmission", by Paul Wessel,
 *    Computers & Geosciences, 29, (5), 665-671, 2003.
 *
 * grdzip takes an short int grid file and encodes it using
 * d^2/dxdy differences which are stored in a packed 8/16/24/32-bit format that varies
 * with the size of the increments.  The packed format is written out via the bzip2
 * compression library routines for further compression.
 *
 * There are 16 formats: 2 use 1-byte encoding, 2 use 2-bytes, 6 use 3-byte encoding, while
 * the rest uses 4-byte encoding.  Below, D is the data increment value, s is signbit, and b
 * is databit(s).  The sign bit allow us to get -D as well as +D. - is an unused bit.  In all
 * cases the first 4 bits are the format bits which tells us which of the 16 encoding formats
 * were used.
 *
 * -----------------------------------------------------------------------------------------
 * ID# flag(4) |      databits(4)             | Store N  D <= max  in  M bits each
 * -----------------------------------------------------------------------------------------
 *   0   0000  | bbbb 			      | Store N = 4-19 zeros. bbbb is (N-4)
 *   1   0001  | bbbb 			      | Store 4 D = 0 or +1 in 1 bits each (1 == +1)
 * -----------------------------------------------------------------------------------------
 * ID# flag(4) |      databits(12)            | Store N  D <= max  in  M bits each
 * -----------------------------------------------------------------------------------------
 *   2   0010  | sbsbsbsbsbsb	 	      | Store 6 D <=         1 in  2 bits each
 *   3   0011  | sbbsbbsbbsbb 		      | Store 4 D <=         3 in  3 bits each
 * -----------------------------------------------------------------------------------------
 * ID# flag(4) |      databits(20)            | Store N  D <= max  in  M bits each
 * -----------------------------------------------------------------------------------------
 *   4  0100   | sbsbsbsbsbsbsbsbsbsb	      | Store 10 D <=         1 in  1 bits each
 *   5  0101   | sbbbsbbbsbbbsbbbsbbb	      | Store  5 D <=         7 in  4 bits each
 *   6  0110   | sbbbbsbbbbsbbbbsbbbb	      | Store  4 D <=        15 in  5 bits each
 *   7  0111   | --sbbbbbsbbbbbsbbbbb         | Store  3 D <=        31 in  6 bits each
 *   8  1000   | sbbbbbbbbbsbbbbbbbbb         | Store  2 D <=       511 in 10 bits each
 *   9  1001   | sbbbbbbbbbbbbbbbbbbb	      | Store  1 D <=    524287 in 20 bits each
 * -----------------------------------------------------------------------------------------
 * ID# flag(4) |      databits(28)            | Store N  D <= max  in  M bits each
 * -----------------------------------------------------------------------------------------
 *  10  1010   | sbsbsbsbsbsbsbsbsbsbsbsbsbsb | Store 14 D <=         1 in  1 bits each
 *  11  1011   | -sbbsbbsbbsbbsbbsbbsbbsbbsbb | Store  9 D <=         3 in  3 bits each
 *  12  1100   | sbbbsbbbsbbbsbbbsbbbsbbbsbbb | Store  7 D <=         7 in  4 bits each
 *  13  1101   | sbbbbbbsbbbbbbsbbbbbbsbbbbbb | Store  4 D <=        63 in  7 bits each
 *  14  1110   | -sbbbbbbbbsbbbbbbbbsbbbbbbbb | Store  3 D <=       255 in  9 bits each
 *  15  1111   | sbbbbbbbbbbbbbsbbbbbbbbbbbbb | Store  2 D <=      8191 in 14 bits each
 * -----------------------------------------------------------------------------------------
 *
 * Values larger than 524287 will be encoded as follows:
 *   (1) write out format 7, and set all 20 data bits to 1; this is a unique flag since
 *       normally the first two databits are unused.
 *   (2) Write the value out using a 4-byte integer directly.
 * 
 * More technical details can be found in the grdzip.h include file.
 * If we compute the ratio bits-per-increment we find which formats offers the best compaction.
 * We will preferentially try to use those formats before trying the others.  The list is:
 *
 * PRIORITY	| ID# | bits-per-increment | Compresses to (%)
 * -----------------------------------------------------
 *	1	|  0  |    0.421 - 1.6	   |	2.6-10	97.4-90 % space savings
 *	2	|  1  |		2.000	   |	12.5	87.5% space savings
 *	3	| 10  |		2.286	   |	14.3	etc.
 *	4	|  4  |		2.400	   |	15.0
 *	5	|  2  |		2.667	   |	16.7
 *	6	| 11  |		3.556	   |	22.2
 *	7	|  3  |		4.000	   |	25
 *	8	| 12  |		4.571	   |	28.6
 *	9	|  5  |		4.800	   |	30.0
 *	10	|  6  |		6.000	   |	37.5
 *	11	| 13  |		8.000	   |	50.0
 *	12	|  7  |		8.000	   |	50.0
 *	13	| 14  |		10.667	   |	66.7
 *	14	|  8  |		12.000	   |	75.0
 *	15	| 15  |		16.000	   |	100.0	No space saving
 *	16	|  9  |		24.000	   |	150.0	Output takes 50% MORE space
 *	17	| 16  |		32.000	   |	150.0	Output takes 50% MORE space
*
 * Therefore, for each incoming string of increments we seek the most efficient
 * format that can store these increments.  We do this by testing each format
 * in the order of the priority established above.
 * We use format X when we find there are N consecutive 2-byte integer
 * data increments that each can be stored using only M bits .  We
 * determine what X is by trying all formats in order of priority and finding
 * the highest priority X that works.  We then pack the bits accordingly and write
 * out 1-4 unsigned bytes.
 * For each original scanline of data values, we begin the processing by
 * first calculating the d2/dxy differences.  Later, when we decode
 * the packed file we keep track of the number of encoded data points
 * so we will know when we have read the last packed word for a given
 * scanline.  The endianness of the encoding platform is also stored.
 * Upon decoding with grdzip -d, the increments are then automatically
 * integrated over rows to arrive at the original data values.  The packed
 * data are read and written via the bzip2 library to achive maximum compression
 * of file size.
 *
 * Geoware, (c) 2000-2012 P. Wessel
 *
 * For bzip2 source code, see www.bzip.org
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 * Revision 1/1/2005:	Added -4 option for 4-byte integer grids
 * Revision 1/1/2006:	Added -1 option for 1-byte integer grids
 * Revision 1/10/2008:	Do not write (empty) output grid if -Q is selected
 * Revision 11/15/2011:	Use LONG counters to handle really large files
 */
 
#include "grdzip.h"

/* Function prototypes; see below for description of purpose */

void do_the_encode (char *infile, char *outfile, BOOLEAN swap_on_input, BOOLEAN has_header, int nx, int ny, int grid_word_size, BOOLEAN verbose);
void do_the_decode (char *infile, char *outfile, char want_endian, BOOLEAN write_header, BOOLEAN verbose, int in_size, BOOLEAN query);
void double_swab (double *x);
int read_GMT_header (BZFILE *bz, struct GRD_HEADER *header, char *program);
int write_GMT_header (BZFILE *bz, struct GRD_HEADER *header, char *program, unsigned int *nWritten);
int find_best_sqeeze (int z[], BOOLEAN s[], int i, int nx, int *n_pass);

int priority[N_FORMATS] = { 0, 1, 10, 4, 2, 11, 3, 12, 5, 6, 13, 7, 14, 8, 15, 9, 16};
int run_mode = 0;	/* Only used to control format of messages */

int main (int argc, char **argv) {

	int i, n_files = 0;			/* Misc. counters */
	int nx = 0, ny = 0;			/* Dimension of input grid if -H is not set */
	int in_size = 0;			/* File size of packed file is read from stdin */
	int grid_word_size = 2;			/* Number of bytes per grid element [2 for short is default] */
	BOOLEAN has_header = FALSE;		/* TRUE if input grid has GMT header structure */
	BOOLEAN write_header = TRUE;		/* TRUE if output grid should contain GMT header structure */
	BOOLEAN swap_on_input = FALSE;		/* TRUE if we must first byte-swap the input data */
	BOOLEAN compress = TRUE;		/* TRUE if we want to compress; FALSE for decompress */
	BOOLEAN verbose = FALSE;		/* TRUE if we want verbose feedback (-V) */
	BOOLEAN error = FALSE;			/* TRUE if incorrect or missing arguments are set */
	BOOLEAN query = FALSE;			/* TRUE if we just want information about compressed file */
	char want_endian = 'H';			/* Desired endianness of unzipped file [same as host] */
	char *etype[2] = {"Little", "Big"};	/* The two types of endian-ness */
	char *infile = NULL;			/* Input file name (NULL for stdin) */
	char *outfile = NULL;			/* Output file name (NULL for stdout) */
	
	for (i = 1; i < argc; i++) {		/* Processes command line arguments */
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			
				/* Cases that apply for decompression only: */

				case 'D':
				case 'd':
					compress = FALSE;
					if (argv[i][2] == 'e') run_mode = 1;	/* Geoware message formatting */
					break;
				case 'e':
				case 'E':
					if (argv[i][2] == 'l' || argv[i][2] == 'L')
						want_endian = 'L';
					else if (argv[i][2] == 'b' || argv[i][2] == 'B')
						want_endian = 'B';
					else if (argv[i][2] == 'o' || argv[i][2] == 'O')
						want_endian = 'O';
					else if (argv[i][2] == '\0' || argv[i][2] == 'h' || argv[i][2] == 'H')
						want_endian = 'H';
					else {
						fprintf (stderr, "grdzip: Error: Choose among -E, -EB, -EL, -EH only\n");
						error = TRUE;
					}
					break;
				case 'i':	/* Use the incoming bzip2 file size instead of packed file size when reporting expansion */
				case 'I':
					in_size = atoi (&argv[i][2]);
					break;
					
				/* Cases that apply for compression only */
				
				case 'C':
				case 'c':
					compress = TRUE;
					break;
				case 'X':
				case 'x':
					nx = atoi (&argv[i][2]);
					break;
				case 'Y':
				case 'y':
					ny = atoi (&argv[i][2]);
					break;
				case 'H':
				case 'h':
					has_header = TRUE;
					break;
				case 'N':
				case 'n':
					write_header = FALSE;
					break;
				case 'Q':
				case 'q':
					query = TRUE;
					break;
				case 'S':
				case 's':
					swap_on_input = TRUE;
					break;
				case '1':	/* Dealing with 1-byte int grid */
					grid_word_size = 1;
					break;
				case '2':	/* Dealing with 2-byte int grid */
					grid_word_size = 2;
					break;
				case '4':	/* Dealing with 4-byte int grid */
					grid_word_size = 4;
					break;
					
				/* Common options : */
				
				case 'f':
				case 'F':
					outfile = &argv[i][2];
					break;
				case 'v':
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else {
			n_files++;
			infile = argv[i];
		}
	}
	
	if (n_files > 1) {
		fprintf (stderr, "grdzip: Error: Only 1 input file (or stdin) allowed on the command line\n");
		exit (EXIT_FAILURE);
	}

	if (argc == 1 || error) {
		fprintf (stderr, "grdzip %s - Compression/Decompression of 1-, 2-, or 4-byte integer data sets\n\n", PROGRAM_VERSION);
		fprintf (stderr, "grdzip will read from infile [stdin] and write to outfile [stdout].\n");
		fprintf (stderr, "grdzip -D will decompress, grdzip [-C] will compress [Default].\n\n");
		fprintf (stderr, "usage (compression):\n\n");
		fprintf (stderr, "grdzip -C [grdfile] [-1|2|4] [-H] [-S] [-V] [-X<nx>] [-Y<ny>] [-F<packedfile>]\n");
		fprintf (stderr, "\t-1 Integer grid is 1-byte ints [default is 2-byte short int]\n");
		fprintf (stderr, "\t-2 Integer grid is 2-byte ints [default]\n");
		fprintf (stderr, "\t-4 Integer grid is 4-byte ints [default is 2-byte short int]\n");
		fprintf (stderr, "\t-F Packed file name [default writes to stdout]\n");
		fprintf (stderr, "\t-H Input file has a GMT grid %d-byte header [default is no header]\n", HEADER_SIZE);
		fprintf (stderr, "\t-S First swap byte-order of input [default is no swap]\n");
		fprintf (stderr, "\t-V Report on progress [default is silent]\n");
		fprintf (stderr, "\t-X set number of columns in grid (only needed if no header)\n");
		fprintf (stderr, "\t-Y set number of rows in grid (only needed if no header)\n\n");
		fprintf (stderr, "usage (decompression):\n\n");
		fprintf (stderr, "grdzip -D [packedfile] [-E<endian>] [-I<filesize>] [-N] [-Q] [-V] [-F<grdfile>]\n");
		fprintf (stderr, "\t-E Desired endianness of the output file.  Choose from\n");
		fprintf (stderr, "\t   -EB : Big endian.\n");
		fprintf (stderr, "\t   -EL : Little endian.\n");
		fprintf (stderr, "\t   -EO : Same as the original file at the time of encoding.\n");
		fprintf (stderr, "\t   -EH or -E : Same as this machine (%s endian) [Default].\n", etype[WORDS_BIGENDIAN]);
		fprintf (stderr, "\t-F Grid file name [default writes to stdout]\n");
		fprintf (stderr, "\t-I Give packed file size when stdin is read (for -V).\n");
		fprintf (stderr, "\t-N Do not write a GMT grid %d-byte header it present [default writes header if present]\n", HEADER_SIZE);
		fprintf (stderr, "\t-Q Just report information about the compressed grid file [Default extracts the grid].\n");
		fprintf (stderr, "\t-V Report on progress [default is silent]\n");
		
		exit (EXIT_FAILURE);
	}
	
	if (compress)
		do_the_encode (infile, outfile, swap_on_input, has_header, nx, ny, grid_word_size, verbose);
	else
		do_the_decode (infile, outfile, want_endian, write_header, verbose, in_size, query);

	exit (EXIT_SUCCESS);
}

/* This is the routine that handles the encoding of the compressed/packed input data */

void do_the_encode (char *infile, char *outfile, BOOLEAN swap_on_input, BOOLEAN has_header, int nx, int ny, int grid_word_size, BOOLEAN verbose)
{
	int n_pass;				/* Number of increments encoded in this string */
	int format;				/* Format id (0-15) for this string */
	int bzerror;				/* Holds error conditions from bzip2 routines */
	int extra;				/* Flag to store the grdzip version */
	int val32;				/* Temp value used to write */
	int *dzdy, *d2zdxdy;			/* Pointers to arrays with dz/dy and d^2z/dxdy increments */
	int i, j, k, ii, kk;			/* Misc. counters */

	unsigned int swap_test;			/* Special word written to help us identify endianness */
	unsigned int word32;			/* 32-bit word used to hold encoded increments */
	unsigned int nRead = 0;			/* bzip2 counters */
	unsigned int nWritten = 0;		/* Number of compressed bytes written */
	ULONG orig_size;			/* Size (in bytes) of input data file */
	ULONG packed_size;			/* Size (in bytes) or compressed output file */
	ULONG count[N_FORMATS];			/* Number of use for each format */
	ULONG words[N_FORMATS];			/* Number of increments stored for each format */

	short int *row_short;			/* Pointer to array with this row's data values, if 16-bit */
	int *row;				/* Pointer to array with this row's data values */
	int *last_row;				/* Pointer to array with last row's data values */
	unsigned char bytes[4];			/* 4 bytes used to write a 32-bit word the slow way */

	BOOLEAN *sign;				/* Pointer to array with increment signs (TRUE is -ve) */

	char endian;				/* Endianness of the computer (L or B) */
	char *row_byte;				/* Pointer to array with this row's data values, if 8-bit */

	struct GRD_HEADER header;		/* Structure used for GMT grd headers */

	FILE *fpi;				/* Input file pointer [stdin] */
	FILE *fpo;				/* Output file pointer [stdout] */
	BZFILE *bz;				/* File pointer for bzip2 library routines */

	if (infile) {	/* Got input file */
		if ((fpi = fopen (infile, "rb")) == NULL) {
			fprintf (stderr, "grdzip: Error opening file %s [not found?]\n", infile);
			exit (EXIT_FAILURE);
		}
		if (verbose) fprintf (stderr, "grdzip: Packing %s ", infile);
	}
	else {		/* Read from stdin instead */
		fpi = stdin;
#if defined(_WIN32) || defined(WIN32)
		_setmode (fileno (fpi), _O_BINARY);
#elif defined(__EMX__)
		_fsetmode (fpi, "b");
#endif		
		if (verbose) fprintf (stderr, "grdzip: Packing %s ", infile);
	}
	if (outfile) {	/* Got an output file name */
		if ((fpo = fopen (outfile, "wb")) == NULL) {
			fprintf (stderr, "grdzip: Error creating file %s [permission problem?]\n", outfile);
			exit (EXIT_FAILURE);
		}
		if (verbose) fprintf (stderr, "to %s\n", outfile);
	}
	else {		/* Write to stdout instead */
		fpo = stdout;
#if defined(_WIN32) || defined(WIN32)
		setmode (fileno (fpo), _O_BINARY);
#elif defined(__EMX__)
		_fsetmode (fpo, "b");
#endif		
		if (verbose) fprintf (stderr, "to <stdout>\n");
	}
	
	if (has_header) {	/* We are dealing with a GMT native binary grid */
		if (fread ((void *)&header.nx, 3*sizeof (int), (size_t)1, fpi) != 1 || fread ((void *)&header.x_min, sizeof (struct GRD_HEADER) - ((long)&header.x_min - (long)&header.nx), (size_t)1, fpi) != 1) {
			fprintf (stderr, "grdzip: Error reading GMT grd header structure\n");
			exit (EXIT_FAILURE);
		}
		nx = header.nx;
		ny = header.ny;
	}
	
	if (nx <= 0 || ny <= 0) {	/* Either no header present or user forgot to give command line options */
		fprintf (stderr, "grdzip: Must set -X and -Y\n");
		exit (EXIT_FAILURE);
	}

	/* Allocate necessary memory for rows and their derivatives */
	
	if (grid_word_size == 1 && (row_byte = (char *) calloc ((size_t)nx, sizeof (char))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for row_byte\n");
		exit (EXIT_FAILURE);
	}
	else if (grid_word_size == 2 && (row_short = (short int *) calloc ((size_t)nx, sizeof (short int))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for row_short\n");
		exit (EXIT_FAILURE);
	}
	if ((row = (int *) calloc ((size_t)nx, sizeof (int))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for row\n");
		exit (EXIT_FAILURE);
	}
	if ((last_row = (int *) calloc ((size_t)nx, sizeof (int))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for last_row\n");
		exit (EXIT_FAILURE);
	}
	if ((dzdy = (int *) calloc ((size_t)nx, sizeof (int))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for dzdy\n");
		exit (EXIT_FAILURE);
	}
	if ((d2zdxdy = (int *) calloc ((size_t)nx, sizeof (int))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for row\n");
		exit (EXIT_FAILURE);
	}
	if ((sign = (BOOLEAN *) calloc ((size_t)nx, sizeof (BOOLEAN))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for row\n");
		exit (EXIT_FAILURE);
	}

	memset ((void *)count, 0, N_FORMATS * sizeof (ULONG));
	memset ((void *)words, 0, N_FORMATS * sizeof (ULONG));
	
#if WORDS_BIGENDIAN == 0
	endian = 'L';	/* Little-endianness */
#else
	endian = 'B';	/* Big-endianness */
#endif

	/* Open output compressed file via bzip2 routines */
	
	bz = BZ2_bzWriteOpen (&bzerror, fpo, 9, 0, 30);	/* Maximum bzip2 compression (9) requested */
	if (bzerror != BZ_OK) {
		fprintf (stderr, "grdzip: Error initializing compressed output\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)&endian, sizeof (char));	/* The encoding machine's endianness */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error writing endian\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}
	swap_test = 65537; /* Binary 00000000|00000001|00000000|00000001 */
	BZ2_bzWrite (&bzerror, bz, (void *)&swap_test, sizeof (int));	/* The swap test for later reading */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error writing swaptest\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)&ny, sizeof (int));	/* Write ny */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error writing ny\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)&nx, sizeof (int));	/* Write nx */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error writing nx\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}
	extra = (has_header | GRDZIP_VERSION);	/* Encode the packed version number so grdzip can do the right thing */
	if (grid_word_size == 4) extra += 128;	/* High bit used to indicate 4-byte grid */
	if (grid_word_size == 1) extra += 64;	/* Next-high bit used to indicate 1-byte grid */
	if (grid_word_size == 1) swap_on_input = FALSE;	/* No swapping can be done - value is ignored */
	bytes[0] = (unsigned char) extra;	/* Only need 1 byte for this */
	BZ2_bzWrite (&bzerror, bz, (void *)bytes, sizeof (unsigned char));	/* Write the combined flags for header and version */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error writing extra\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}

	if (has_header && write_GMT_header (bz, &header, "grdzip", &nWritten)) {	/* Write the GMT header */
		fprintf (stderr, "grdzip: Error writing GMT grd header\n");
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
		exit (EXIT_FAILURE);
	}
	
	/* Start processing input data rows.  The initial last_row has zeros only, so dzdy = row for j = 0 */
	
	packed_size = 0;
	for (j = 0; j < ny; j++) {
		if (verbose && (j%100 == 0)) fprintf (stderr, "grdzip: Working on row %6.6d\r", j);
		
		memcpy ((void *)last_row, (void *)row, (size_t)(nx * sizeof (int)));	/* Make current row the previous row */

		if (grid_word_size == 1) {
			if (fread ((void *)row_byte, sizeof(char), nx, fpi) != (size_t)nx) {		/* Get the current row values */
				fprintf (stderr, "grdzip: Error reading byte row %d\n", j);
				exit (EXIT_FAILURE);
			}
			for (i = 0; i < nx; i++) row[i] = (int)row_byte[i];
		}
		else if (grid_word_size == 2) {
			if (fread ((void *)row_short, sizeof(short int), nx, fpi) != (size_t)nx) {	/* Get the current row values */
				fprintf (stderr, "grdzip: Error reading short row %d\n", j);
				exit (EXIT_FAILURE);
			}
			for (i = 0; i < nx; i++) row[i] = (int)row_short[i];
		}
		else {
			if (fread ((void *)row, sizeof(int), nx, fpi) != (size_t)nx) {		/* Get the current row values */
				fprintf (stderr, "grdzip: Error reading int row %d\n", j);
				exit (EXIT_FAILURE);
			}
		}
		
		if (swap_on_input) row[0] = ((grid_word_size == 2) ? swab_2 (row[0]) : swab_4 (row[0]));		/* Col 0: Take care of byte swapping */
		d2zdxdy[0] = dzdy[0] = row[0] - last_row[0];		/* Col 0: Do d/dy ; also keep value in dz2dxdy[0] */
		sign[0] = (d2zdxdy[0] < 0);				/* Col 0: Keep track of increment signs */
		if (sign[0]) d2zdxdy[0] = -d2zdxdy[0];			/* Col 0: Only keep absolute values here */
		for (i = 1; i < nx; i++) {				/* For the remaining columns: */
			if (swap_on_input) row[i] = ((grid_word_size == 2) ? swab_2 (row[i]) : swab_4 (row[i]));		/* Take care of byte swapping */
			dzdy[i] = row[i] - last_row[i];				/* Do d/dy */
			d2zdxdy[i] = dzdy[i] - dzdy[i-1];			/* Do d2/dxdy */
			sign[i] = (d2zdxdy[i] < 0);				/* Keep track of increment signs */
			if (sign[i]) d2zdxdy[i] = -d2zdxdy[i];			/* Only keep absolute values here */
		}
		for (i = 0; i < nx; i += n_pass) {	/* Process this row for packing */

			/* Try the various formats in order of best storage-saving property */
			
			format = find_best_sqeeze (d2zdxdy, sign, i, nx, &n_pass);	/* Returns format id as well as n_pass, the # of values to store */
			
			if (format == 16) {	/* Special case for a value too large for the general scheme below */
				bytes[0] = 127;
				BZ2_bzWrite (&bzerror, bz, (void *)bytes, sizeof (unsigned char));	/* Write unique flag...*/
				val32 = (sign[i]) ? -d2zdxdy[i] : d2zdxdy[i];				/* ...restore the sign, if any,... */
				BZ2_bzWrite (&bzerror, bz, (void *)&val32, sizeof (int));		/* ...and write the 4 byte int directly */
			}
			else if (format == 0) {	/* OK, save special 0/+1 increment byte */
				bytes[0] = (unsigned char)(n_pass - 4);					/* so 0-15 means 4-19 consecutive zeros */
				BZ2_bzWrite (&bzerror, bz, (void *)bytes, sizeof (unsigned char));	/* Write single byte word */
			}
			else if (format == 1) {	/* Run of 4 0 or +1 */					
				word32 = template[format].format_id;		/* Highest 4-bits hold the format id */
				ii = i;
				if (d2zdxdy[ii++]) word32 |= 8;	/* +1 is ON bit, 0 is OFF bits */
				if (d2zdxdy[ii++]) word32 |= 4;
				if (d2zdxdy[ii++]) word32 |= 2;
				if (d2zdxdy[ii++]) word32 |= 1;
				bytes[0] = (unsigned char)word32;
				BZ2_bzWrite (&bzerror, bz, (void *)bytes, sizeof (unsigned char));	/* Write single byte word */
			}
			else {	/* Use the 16/24/32 bit approach */
				word32 = d2zdxdy[i];						/* Encode the first increment abs value */
				if (sign[i]) word32 += template[format].sign_bit;		/* Encode the first increment sign bit if needed */
				for (ii = i + 1, kk = 1; kk < n_pass; kk++, ii++) {		/* For each additional increment to be encoded */
					word32 <<= template[format].bits;			/* Move previously encoded increment to the left */
					word32 += d2zdxdy[ii];					/* Add in the current increment */
					if (sign[ii]) word32 += template[format].sign_bit;	/* Encode the sign bit if needed */
				}
				word32 |= template[format].format_id;				/* Add format identifier */
				/* Encode the 16/24/32 bit integer as 2/3/4 bytes */
				for (k = 0; k < (int)template[format].n_bytes; k++) {
					bytes[3-k] = (unsigned char) (word32 & 255);		/* Rightmost 8 bits */
					word32 >>= 8;						/* Chop off rightmost bits */
				}
				BZ2_bzWrite (&bzerror, bz, (void *)&bytes[4-k], (size_t)(template[format].n_bytes * sizeof (unsigned char)));	/* Write 2-4 bytes */
				if (bzerror == BZ_IO_ERROR) {
					fprintf (stderr, "grdzip: Error writing word32\n");
					BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
					exit (EXIT_FAILURE);
				}
			}
			packed_size += template[format].n_bytes * sizeof (unsigned char);	/* This is size of packed file */
			count[format] += n_pass;						/* Keep track of # of increments written in each format */
			words[format]++;							/* Keep track of format usage */
		}
	}
	
	/* Done with all data rows; close compressed output stream */
	
	BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, &nWritten);
	if (bzerror != BZ_OK) {
		fprintf (stderr, "grdzip: Error closing compressed file\n");
		exit (EXIT_FAILURE);
	}
	if (fpi != stdin) fclose (fpi);	/* Close input unless it is stdin */
	
	if (verbose) fprintf (stderr, "grdzip: Working on row %6.6d\n", j);	/* To get the linefeed for last message */

	orig_size = ((ULONG)grid_word_size) * ((ULONG)nx) * ((ULONG)ny);	/* Full size of 1|2|4-byte grid without any header */
	packed_size += 14 ;					/* Add 14 header bytes + 4 bytes per row to output size */
	if (has_header) packed_size += HEADER_SIZE, orig_size += HEADER_SIZE;	/* Also add size of GMT header, if present */
	if (verbose) {
		fprintf (stderr, "format |   words   | increments\n");
		fprintf (stderr, "-------------------------------\n");
		for (i = 0; i < N_FORMATS; i++) fprintf (stderr, "  %2d   |%12lu |%14lu \n", i, words[i], count[i]);
		fprintf (stderr, "-------------------------------\n");
		fprintf (stderr, "\nOriginal file size (bytes): %14lu [100.00 %%]\n", orig_size);
		fprintf (stderr, "Packed file size   (bytes): %14lu [%6.2f %%]\n", packed_size, (100.0 * packed_size) / orig_size);
		fprintf (stderr, "Final grdzip size  (bytes): %14u [%6.2f %%]\n", nWritten, (100.0 * nWritten) / orig_size);
	}
	
	/* Free up the memory used */
	
	free ((void *)row);
	free ((void *)last_row);
	free ((void *)d2zdxdy);
	free ((void *)dzdy);
	free ((void *)sign);
	if (grid_word_size == 1) free ((void *)row_byte);
	if (grid_word_size == 2) free ((void *)row_short);
}

int write_GMT_header (BZFILE *bz, struct GRD_HEADER *header, char *program, unsigned int *nWritten)
{
	/* Write GRD header structure to native binary file via bzip2 lib.
	 * Because internal arrangement of structure members are platform
	 * dependent, we write the information using arrays. */

	int n[3];			/* Array used for nx, ny, and node_offset */
	int bzerror;			/* Holds error codes returned from bzip2 lib */
	unsigned int nRead = 0;		/* Number of uncompressed bytes read */
	double d[10];			/* Array used for the 10 double precision items in the GMT grd header */
	
	/* Copy structure members to arrays  */
	
	n[0] = header->nx;	n[1] = header->ny;	n[2] = header->node_offset;
	d[0] = header->x_min;	d[1] = header->x_max;	d[6] = header->x_inc;
	d[2] = header->y_min;	d[3] = header->y_max;	d[7] = header->y_inc;
	d[4] = header->z_min;	d[5] = header->z_max;
	d[8] = header->z_scale_factor;	d[9] = header->z_add_offset;
	
	/* Write GRD header via these temporary arrays */
	
	BZ2_bzWrite (&bzerror, bz, (void *)n, 3*sizeof (int));		/* Write nx, ny, and node_offset */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing nx, ny, and node_offset to grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)d, 10*sizeof (double));	/* Write w/e/s/n/z0/z1/dx/dy/zs/zo */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing w/e/s/n/z0/z1/dx/dy/zs/zo from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	
	/* Write text members directly */
	
	BZ2_bzWrite (&bzerror, bz, (void *)header->x_units, GRD_UNIT_LEN*sizeof (char));	/* Write x_unit */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing x_unit text from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)header->y_units, GRD_UNIT_LEN*sizeof (char));	/* Write y_unit */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing y_unit text from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)header->z_units, GRD_UNIT_LEN*sizeof (char));	/* Write z_unit */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing z_unit text from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)header->title, GRD_TITLE_LEN*sizeof (char));	/* Write title */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing title text from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)header->command, GRD_COMMAND_LEN*sizeof (char));	/* Write command */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing command text from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}
	BZ2_bzWrite (&bzerror, bz, (void *)header->remark, GRD_REMARK_LEN*sizeof (char));	/* Write remark */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error writing remark text from grdfile header!\n", program);
		BZ2_bzWriteClose (&bzerror, bz, 0, &nRead, nWritten);
		return (TRUE);
	}

	return (FALSE);
}

int find_best_sqeeze (int z[], BOOLEAN s[], int i, int nx, int *n_pass)
{
	int ii, kk, k, stop, format, this;

	/* find_best_squeeze will determine which format is best suited to store the next
	 * string of increments.  We search for the appropriate format that is most efficient */
	 
	/* Check if we have a value too large for the general scheme */
	
	if (abs (z[i]) > 524287) {	/* Too large, must encode as is via special flagging */
		*n_pass = 1;
		return (16);
	}
		
	/* First see if we have a run of 4-19 zeros */
	
	ii = i, (*n_pass) = 0;
	while (z[ii] == 0 && ii < nx && (*n_pass) < (int)template[0].n_in_row) (*n_pass)++, ii++;
	if ((*n_pass) > 3) return (0);	/* 4 or more is great */
	
	/* First see if we have a run of 4 +1/0 */
	
	ii = i, (*n_pass) = 0;
	while (s[ii] == 0 && z[ii] <= 1 && ii < nx && (*n_pass) < (int)template[1].n_in_row) (*n_pass)++, ii++;
	if ((*n_pass) == 4) return (1);	/* 4 is required */
	
	/* None of the 8-bit formats worked; Search for the 16/24/32-bit format that best fits */
	
	for (k = 2, stop = 0; !stop && k < N_FORMATS; k++) {	/* Search for optimal format in order of decreasing efficiency */
		this = priority[k];
		for (ii = i, kk = (*n_pass) = 0; (*n_pass) == kk && ii < nx && kk < (int)template[this].n_in_row; ii++, kk++) {
			if (z[ii] <= (int)template[this].size) (*n_pass)++;
		}
		if ((*n_pass) == (int)template[this].n_in_row) stop = 1, format = this;
	}
	
	if (stop == 0) {
		fprintf (stderr, "grdzip: Fatal Error: Failed to determine a suitable format!\n");
		exit (EXIT_FAILURE);
	}
	return (format);
}

void do_the_decode (char *infile, char *outfile, char want_endian, BOOLEAN write_header, BOOLEAN verbose, int in_size, BOOLEAN query)
{
	double in_Mb, out_Mb;		/* Size of input/output in Mb */
	double pack;			/* Packing ratio */
	double two_percent, last_out;	/* Misc. counters */

	int format;			/* format of the current encoded string */
	int nx, ny;			/* Number of columns and rows in original grid */
	int extra;			/* Combo flag for GMT grd header present and encoding version */
	int n_out = 0, n_in = 0;	/* size counters (in bytes) */
	int VAL;			/* Full 4-byte integers */
	int ii, i, j, k, m_len, width, unit = 1;		/* Misc. counters */
	int delta;			/* Single increment */
	int sign;			/* sign of increment */
	int out_sign;			/* Current progress bar indicator symbol to print */
	int bzerror;			/* Error indicator for bzip2 */
	int grid_word_size = 2;		/* Size of integer used in grid (1|2|4) [2]*/

	unsigned int swap_test;		/* Test integer to see if swapping is needed on input */
	unsigned int word32;		/* Our 32-bit word */

	int value[32];			/* String of current increments */
	int val32, int4;		/* Misc. 4-byte values */
	int *last_row;			/* Array of last row (needed to integrate) */

	short int val16, int2;		/* Misc. 2-byte values */

	unsigned char byte;		/* Single byte to read from compressed input */
	char out_marker[51] = "----1----2----3----4----5----6----7----8----9----|";	/* Progress bar */
	char endian;			/* Endianness of the system that wrote the compressed file */
	char name[BUFSIZ];		/* Name of file */
	char message[BUFSIZ];		/* Information message */
	char *size_name[5] = {"", "byte", "short", "", "int"};
	char *prefix_msg[2] = {"grdzip: Unpacking", "-->"};
	char *YN = "NY", *u[3] = {"Gb", "Mb", "kb"};
	char val8;
#if defined(_WIN32) || WORDS_BIGENDIAN == 0
	char my_endian = 'L';		/* Current system is Little-endian */
#else
	char my_endian = 'B';		/* Current system is Big-endian */
#endif

	BOOLEAN swap_on_input;		/* TRUE if we must swap integers read from compressed input */
	BOOLEAN swap_on_output;		/* TRUE if we should swap the output data */
	BOOLEAN has_header;		/* TRUE if we have a GMT grd header present */

	FILE *fpi, *fpo;		/* Input and output file pointers */
	BZFILE *bz;			/* File pointer for compressed input */

	struct GRD_HEADER header;	/* GMT Header structure */
	struct GRD_HEADER raw_header;	/* Same, just as it was read */

	memset ((void *)message, 0, BUFSIZ);
	if (infile) {	/* Got a file name */
		if (in_size == 0) {	/* Get incoming file size */
			struct STAT buf;
			STAT (infile, &buf);
			in_size = buf.st_size;
		}
		if ((fpi = fopen (infile, "rb")) == NULL) {
			fprintf (stderr, "grdzip: Error opening file %s [not found?]\n", infile);
			exit (EXIT_FAILURE);
		}
		strcpy (name, infile);
	}
	else {
		fpi = stdin;
#if defined(_WIN32) || defined(WIN32)
		setmode (fileno (fpi), _O_BINARY);
#elif defined(__EMX__)
		_fsetmode (fpi, "b");
#endif		
		strcpy (name, "<stdin>");
	}
	sprintf (message, "%s %s to ", prefix_msg[run_mode], name);
	if (verbose) fprintf (stderr, "%s", message);
	m_len = strlen (message);
	memset ((void *)message, 0, BUFSIZ);
	if (outfile) {	/* Got an output file name */
		int start, skip;
		if (!query && (fpo = fopen (outfile, "wb")) == NULL) {
			fprintf (stderr, "grdzip: Error creating file %s [permission problem?]\n", outfile);
			exit (EXIT_FAILURE);
		}
		sprintf (message, "%s", outfile);
		width = 80 - m_len;
		if ((m_len = strlen (message)) > width) {	/* Must shorten message by using ellipses in the middle */
			skip = m_len - width + 3;
			i = start = m_len/2 - skip/2;
			for (k = 0; k < 3; k++, i++) message[i] = '.';	/* Ellipses ... */
			k = start + skip;
			while (message[k]) message[i++] = message[k++];
			message[i] = '\0';
		}
	}
	else {
		if (!query) {
			fpo = stdout;
#if defined(_WIN32) || defined(WIN32)
			setmode (fileno (fpo), _O_BINARY);
#elif defined(__EMX__)
			_fsetmode (fpo, "b");
#endif		
		}
		strcpy (message, "<stdout>");
	}
	if (verbose) fprintf (stderr, "%s\n", message);
	
	/* Initiate reading of compressed input file */
	
	bz = BZ2_bzReadOpen (&bzerror, fpi, 0, 0, NULL, 0);
	if (bzerror != BZ_OK) {
		fprintf (stderr, "grdzip: Error initializing compressed input\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)&endian, sizeof (char));	/* The encoding machine's endianness */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error reading endian\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	
	/* Given past and present endianness, determine if swapping is required */
	
	switch (want_endian) {
		case 'B':	/* Want output to be big-endian */
		case 'L':	/* Want output to be little-endian */
			swap_on_output = (my_endian != want_endian);	/* Must swap if they are different */
			break;
		case 'O':	/* Want to keep original endianness on originating machine */
			swap_on_output = (my_endian != endian);		/* Swap is not same as before */
			break;
		case 'H':	/* want native endianness - do nothing */
			swap_on_output = FALSE;				/* Just ignore the issue */
			break;
	}
	
	BZ2_bzRead (&bzerror, bz, (void *)&swap_test, sizeof (int));	/* The swaptest integer */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error reading swaptest\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	swap_on_input = (swap_test != 65537);			/* If we cannot understand swap_test we need to swap input data */
	
	BZ2_bzRead (&bzerror, bz, (void *)&ny, sizeof (int));	/* ny - the number of rows */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error reading ny\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)&nx, sizeof (int));	/* nx - the number of columns */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error reading nx\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)&byte, sizeof (unsigned char));	/* has_header combo flag (header | version) */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "grdzip: Error reading extra flag\n");
		BZ2_bzReadClose (&bzerror, bz);
		exit (EXIT_FAILURE);
	}
	if (swap_on_input) {
		ny = swab_4 (ny);
		nx = swab_4 (nx);
	}
	extra = (unsigned int)byte;
	has_header = (extra & 1);				/* Take off the version flag */
	if (extra & 128) grid_word_size = 4;			/* Dealing with a 4-byte grid */
	if (extra & 64) grid_word_size = 1;			/* Dealing with a 1-byte grid */
	if (query) {	/* Just tell what we know and exit */
		printf ("grdzip: Name = %s size = %s nx = %d ny = %d header = %c\n", name, size_name[grid_word_size], nx, ny, YN[(int)has_header]);
		exit (EXIT_SUCCESS);
	}
	if ((last_row = (int *) calloc ((size_t)nx, sizeof (int))) == NULL) {
		fprintf (stderr, "grdzip: Could not allocate memory for last_row\n");
		exit (EXIT_FAILURE);
	}
	
	if (has_header && read_GMT_header (bz, &header, "grdzip")) {	/* Must read the compressed GMT grd header */
		fprintf (stderr, "grdzip: Error reading header\n");
		exit (EXIT_FAILURE);
	}
	
	if (has_header && write_header) {	/* Given header data, fill out GMT grd header structure for output */
		raw_header = header;
		if (swap_on_input) {	/* Must swap items in structure on input but never on output */
			header.nx = swab_4 (header.nx);
			header.ny = swab_4 (header.ny);
			header.node_offset = swab_4 (header.node_offset);
			double_swab (&header.x_min);
			double_swab (&header.x_max);
			double_swab (&header.y_min);
			double_swab (&header.y_max);
			double_swab (&header.z_min);
			double_swab (&header.z_max);
			double_swab (&header.x_inc);
			double_swab (&header.y_inc);
			double_swab (&header.z_scale_factor);
			double_swab (&header.z_add_offset);
		}
		if (swap_on_input && swap_on_output) {
			if (fwrite ((void *)&raw_header.nx, 3*sizeof (int), (size_t)1, fpo) != 1 || fwrite ((void *)&raw_header.x_min, sizeof (struct GRD_HEADER) - ((long)&raw_header.x_min - (long)&raw_header.nx), (size_t)1, fpo) != 1) {
				fprintf (stderr, "grdzip: Error writing header\n");
				exit (EXIT_FAILURE);
			}
		}
		else {
			if (fwrite ((void *)&header.nx, 3*sizeof (int), (size_t)1, fpo) != 1 || fwrite ((void *)&header.x_min, sizeof (struct GRD_HEADER) - ((long)&header.x_min - (long)&header.nx), (size_t)1, fpo) != 1) {
				fprintf (stderr, "grdzip: Error writing header\n");
				exit (EXIT_FAILURE);
			}
		}
	}
	
	two_percent = 0.02 * nx * ny;	/* Update progress bar every increment of 2% file size */
	out_sign = 0;
	last_out = 0.0;
	if (verbose) fprintf (stderr, "|");

	for (j = 0; j < ny; j++) {	/* We still process in a row-by-row mode since each row is packed individually */
			
		/* First value difference stored as 4-byte increment */
		VAL = 0;		/* Fresh start for this row */
		i = ii = 0;
		while (ii < nx) {	/* Process this row for unpacking */
			BZ2_bzRead (&bzerror, bz, (void *)&byte, sizeof (unsigned char));	/* first byte */
			if (bzerror == BZ_IO_ERROR) {
				fprintf (stderr, "grdzip: Error reading single byte\n");
				BZ2_bzReadClose (&bzerror, bz);
				exit (EXIT_FAILURE);
			}
			word32 = (unsigned int)byte;				/* Start building the 8-, 24-, or 32-bit word */
			if (word32 == 127) 	/* Special flag for a single 32-bit value to follow */
				format = 16;
			else {
				format = (word32 >> 4);					/* Determine encoding format */
				word32 &= 15;						/* Keep 4 low data bits -> 15 = (1 << 4) - 1*/
			}

			if (format > 16) {
				fprintf (stderr, "grdzip: Row error for j = %d (bad format %d)\n", j, format);
				exit (EXIT_FAILURE);
			}
			
			switch (format) {	/* Take different actions depending on encoding format */
			
				case 0:	/* consecutive zeros format */
					/* The number of zeros are word32 + 4 since 0-15 means 4-19 zeros */
					for (k = 0; k < (int)(word32 + 4); k++) value[template[format].n_in_row-k-1] = 0;
					break;
				case 1:	/* 0 or +1 format */
					k = 0;
					value[k++] = (word32 & 8) ? 1 : 0;	/* value will be sorted in correct order by going from High to Low bit */
					value[k++] = (word32 & 4) ? 1 : 0;
					value[k++] = (word32 & 2) ? 1 : 0;
					value[k++] = (word32 & 1) ? 1 : 0;
					break;
				case 16:	/* The extra large value (one at the time)*/
					BZ2_bzRead (&bzerror, bz, (void *)&word32, sizeof (int));	/* A 4-byte signed integer */
					value[0] = (swap_on_input) ? swab_4 (word32) : word32;		/* Since it was written as an int, not 4 single bytes */
					k = 1;
					break;
				default:	/* One of the 16-, 24- or 32-bit formats */
					for (k = 1; k < (int)template[format].n_bytes; k++) {	/* Read 1, 2, or 3 extra bytes and build the 32-bit word */
						BZ2_bzRead (&bzerror, bz, (void *)&byte, sizeof (unsigned char));	/* 2nd byte */
						word32 <<= 8;								/* Shift previous bits left.. */
						word32 += (unsigned int)byte;						/* ...and stick in the 2nd byte */
					}
					for (k = 0, i = ii; i < nx && k < (int)template[format].n_in_row; k++) {	/* For all packed values (but watch for end of line) */
						delta = word32 & template[format].size;			/* Just get the relevant data bits (no sign) */
						sign = word32 & template[format].sign_bit;		/* TRUE if sign was - */
						if (sign) delta = -delta;				/* OK, restore sign, if any */
						value[template[format].n_in_row-k-1] = delta;		/* Must store backwards given the unpacking order */
						word32 >>= template[format].bits;			/* Move these bits off to the right */
						i++;
					}
			}
			n_in += (int) template[format].n_bytes;
			
			/* Write out the values we got.  Normally k will end up being 0 but at the last string of the row it may not be */
			
			for (k = template[format].n_in_row - k; k < (int)template[format].n_in_row; k++, ii++) {
				/* Must integrate over y as we go */
				VAL += value[k];	/* This restores dzdy[ii] */
				if (ii >= nx) {		/* Bad news - junk input or bug? */
					fprintf (stderr, "grdzip: Row error for j = %d (ii >= nx)\n", j);
					exit (EXIT_FAILURE);
				}
				val32 = VAL + last_row[ii];	/* This integrates to give z[ii] */
				last_row[ii] = val32;		/* We need that next time around */
				if (grid_word_size == 1) {	/* 1-byte short in grid */
					val8 = (char)val32;
					fwrite ((void *)&val8, sizeof(char), 1, fpo);
				}
				else if (grid_word_size == 2) {	/* 2-byte short in grid */
					val16 = (short int)val32;
					if (swap_on_output) {				/* Other endianness requested */
						int2 = swab_2 (val16);
						fwrite ((void *)&int2, sizeof(short int), 1, fpo);
					}
					else
						fwrite ((void *)&val16, sizeof(short int), 1, fpo);
				}
				else {				/* 4-byte int grid */
					if (swap_on_output) {				/* Other endianness requested */
						int4 = swab_4 (val32);
						fwrite ((void *)&int4, sizeof(int), 1, fpo);
					}
					else
						fwrite ((void *)&val32, sizeof(int), 1, fpo);
				}
				n_out++;					/* One more 2|4-byte integer written */
			}
			if (verbose) {						/* If verbose, update progress bar every 2 % */
				while (n_out > (int) (last_out)) {
					fprintf (stderr, "%c", out_marker[out_sign]);
					last_out += two_percent;
					out_sign++;
				}
			}
		}
	}
	
	/* Close the compressed input stream */
	BZ2_bzReadClose (&bzerror, bz);
	if (bzerror != BZ_OK) {
		fprintf (stderr, "grdzip: Error closing compressed file\n");
		exit (EXIT_FAILURE);
	}
	if (fpi != stdin)  fclose (fpi);
	if (fpo != stdout) fclose (fpo);
	if (verbose) {		/* Finish flushing the progress bar indicator */
		while (n_out > (int) (last_out + 0.5) && out_sign < 50) {
			fprintf (stderr, "%c", out_marker[out_sign]);
			last_out += two_percent;
			out_sign++;
		}
	}
	
	n_out *= grid_word_size;			/* Scale size since we have 1|2|4-byte data */
	if (has_header) n_out += HEADER_SIZE;		/* And add in the header size */
	out_Mb = (double)n_out / (1024.0 * 1024.0);	/* Output file size in Mb */
	/* Input size in Mb; either predetermined from file size or calculated here */
	in_Mb = ((in_size) ? in_size : (n_in + 14 + ((has_header) ? HEADER_SIZE : 0))) / (1024.0 * 1024.0);
	if (out_Mb < 1.0) {	/* Switch to kb */
		out_Mb *= 1024.0;
		in_Mb *= 1024.0;
		unit = 2;
	}
	else if (out_Mb > 1024.0) {	/* Switch to Gb */
		out_Mb /= 1024.0;
		in_Mb /= 1024.0;
		unit = 0;
	}
	pack = 100.0 * in_Mb / out_Mb;	/* Compression achieved in precent of original file size */
	
	if (verbose) fprintf (stderr, "%6.1f -> %6.1f %s [%4.1f %%]\n", in_Mb, out_Mb, u[unit], pack);
}

void double_swab (double *x)
{	/* byte swap a double precision value */
	unsigned int *p, save;
	
	p = (unsigned int *)x;				/* Now, p[0] and p[1] are the two 4-byte words */
	save = p[0];	p[0] = p[1];	p[1] = save;	/* Exchange the two words */
	p[0] = swab_4 (p[0]);	p[1] = swab_4 (p[1]);	/* Swab the byte order, and x is now swabed too */
}

int read_GMT_header (BZFILE *bz, struct GRD_HEADER *header, char *program)
{
	/* Read GRD header structure from bzip2 compressed native binary file. */

	int n[3], bzerror;
	double d[10];
	
	/* Read GRD structure int/double members into arrays first */
	
	BZ2_bzRead (&bzerror, bz, (void *)n, 3*sizeof (int));	/* nx, ny, and node_offset */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading nx, ny, and node_offset from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)d, 10*sizeof (double));	/* w/e/s/n/z0/z1/dx/dy/zs/zo */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading w/e/s/n/z0/z1/dx/dy/zs/zo from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	
	/* Read text strings directly into structure */
	
	BZ2_bzRead (&bzerror, bz, (void *)header->x_units, GRD_UNIT_LEN*sizeof (char));	/* x_unit */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading x_unit text from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)header->y_units, GRD_UNIT_LEN*sizeof (char));	/* y_units */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading y_units text from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)header->z_units, GRD_UNIT_LEN*sizeof (char));	/* z_units */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading z_units text from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)header->title, GRD_TITLE_LEN*sizeof (char));	/* title */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading title text from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)header->command, GRD_COMMAND_LEN*sizeof (char));	/* command */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading command text from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}
	BZ2_bzRead (&bzerror, bz, (void *)header->remark, GRD_REMARK_LEN*sizeof (char));	/* remark */
	if (bzerror == BZ_IO_ERROR) {
		fprintf (stderr, "%s: Fatal Error: Error reading remark text from grdfile header!\n", program);
		BZ2_bzReadClose (&bzerror, bz);
		return (TRUE);
	}

	/* Copy info to the remaining structure members */
	
	header->nx = n[0];	header->ny = n[1];	header->node_offset = n[2];
	header->x_min = d[0];	header->x_max = d[1];	header->x_inc = d[6];
	header->y_min = d[2];	header->y_max = d[3];	header->y_inc = d[7];
	header->z_min = d[4];	header->z_max = d[5];
	header->z_scale_factor = d[8];	header->z_add_offset = d[9];
	
	return (FALSE);
}
